""" This module contains functions for loading data. """
from typing import Tuple

import arff
import pandas as pd


def arff_to_pandas(file_path: str) -> pd.DataFrame:
    """ Load data from the ARFF file into a pd.DataFrame.

    Parameters
    ----------
    file_path: str
        Path of the ARFF file

    Returns
    -------
    pandas.DataFrame
        A dataframe of the data in the ARFF file,
        with categorical columns having category dtype.
    """
    with open(file_path, 'r') as arff_file:
        arff_dict = arff.load(arff_file)

    attribute_names, data_types = zip(*arff_dict['attributes'])
    data = pd.DataFrame(arff_dict['data'], columns=attribute_names)
    for attribute_name, dtype in arff_dict['attributes']:
        # 'real' and 'numeric' are probably interpreted correctly, date support needs to be added.
        if isinstance(dtype, list):
            data[attribute_name] = data[attribute_name].astype('category')
    return data


def X_y_from_arff(file_path: str, split_column: str = 'last') -> Tuple[pd.DataFrame, pd.Series]:
    """ Load data from the ARFF file into pandas DataFrame and specified column to pd.Series. "

    Parameters
    ----------
    file_path: str
        path to the ARFF file.
    split_column: str (default='last')
        Column to split and return separately (e.g. target column).
        Value should either match a column name or 'last'.
        If 'last' is specified, the last column is returned separately.

    Returns
    -------
    Tuple[pd.DataFrame, pd.Series]
        Features (everything except split_column) and targets (split_column).
    """
    data = arff_to_pandas(file_path)

    if split_column == 'last':
        return data.iloc[:, :-1], data.iloc[:, -1]
    elif split_column in data.columns:
        return data.loc[:, data.columns != split_column], data.loc[:, split_column]
    else:
        raise ValueError("No column with name {} found in ARFF file {}".format(split_column, file_path))



import uuid
from typing import List, Callable, Optional
from .primitive_node import PrimitiveNode
from .terminal import DATA_TERMINAL, Terminal


class Individual:
    """ A collection of PrimitiveNodes which together specify a machine learning pipeline.

    Parameters
    ----------
    main_node: PrimitiveNode
        The first node of the individual (the estimator node).
    to_pipeline: Callable, optional (default=None)
         A function which can convert this individual into a machine learning pipeline.
         If not provided, the `pipeline` property will be unavailable.
    """

    def __init__(self, main_node: PrimitiveNode, to_pipeline: Optional[Callable] = None):
        self.fitness = None
        self.main_node = main_node
        self._id = uuid.uuid4()
        self._to_pipeline = to_pipeline

    def __eq__(self, other):
        return isinstance(other, Individual) and other._id == self._id

    def __str__(self):
        return """Individual {}\nPipeline: {}\nFitness: {}""".format(self._id, self.pipeline_str(), self.fitness)

    @property
    def pipeline(self):
        """ Calls the `to_pipeline` method on itself. Typically a scikit-learn Pipeline. """
        if self._to_pipeline is None:
            raise AttributeError("`pipeline` not available because `to_pipeline` was not set on __init__.")
        return self._to_pipeline(self)

    @property
    def short_name(self):
        """ str: e.g. "Binarizer>BernoulliNB" """
        return '>'.join([str(primitive._primitive) for primitive in reversed(self.primitives)])

    def pipeline_str(self):
        """ str: e.g. "BernoulliNB(Binarizer(data, Binarizer.threshold=0.6), BernoulliNB.alpha=1.0)" """
        return str(self.main_node)

    @property
    def primitives(self) -> List[PrimitiveNode]:
        """ List[PrimitiveNode]: lists all primitive nodes, starting with the Individual's main node. """
        primitives = [self.main_node]
        current_node = self.main_node._data_node
        while current_node != DATA_TERMINAL:
            primitives.append(current_node)
            current_node = current_node._data_node
        return primitives

    @property
    def terminals(self) -> List[Terminal]:
        """ List[Terminal]: lists all terminals connected to the Individual's primitive nodes. """
        return [terminal for primitive in self.primitives for terminal in primitive._terminals]

    def replace_terminal(self, position: int, new_terminal: Terminal):
        """ Replace the terminal at `position` by `new_terminal` in-place.

        Parameters
        ----------
        position: int
            Position (in self.terminals) of the terminal to be replaced.
        new_terminal: Terminal
            The new terminal to replace the old one with. Must share output type.
        """
        scan_position = 0
        for primitive in self.primitives:
            if scan_position + len(primitive._terminals) > position:
                terminal_to_be_replaced = primitive._terminals[position - scan_position]
                if terminal_to_be_replaced.identifier == new_terminal.identifier:
                    primitive._terminals[position - scan_position] = new_terminal
                    return
                else:
                    raise ValueError("New terminal does not share output type with the one at position {}."
                                     "Old: {}. New: {}.".format(position,
                                                                terminal_to_be_replaced.identifier,
                                                                new_terminal.identifier))
            else:
                scan_position += len(primitive._terminals)
        if scan_position < position:
            raise ValueError("Position {} is out of range with {} terminals.".format(position, scan_position))

    def replace_primitive(self, position: int, new_primitive: PrimitiveNode):
        """ Replace the PrimitiveNode at `position` by `new_primitive`.

        Parameters
        ----------
        position: int
            Position (in self.primitives) of the PrimitiveNode to be replaced.
        new_primitive: PrimitiveNode
            The new PrimitiveNode to replace the old one with. Must share output type.
        """
        last_primitive = None
        for i, primitive_node in enumerate(self.primitives):
            if i == position:
                if primitive_node._primitive.output != new_primitive._primitive.output:
                    raise ValueError("New primitive does not produce same output as the primitive to be replaced.")
                if isinstance(primitive_node._data_node, str):
                    new_primitive._data_node = primitive_node._data_node
                else:
                    new_primitive._data_node = primitive_node._data_node.copy()
                break
            else:
                last_primitive = primitive_node

        if position == 0:
            self.main_node = new_primitive
        else:
            last_primitive._data_node = new_primitive

    def copy_as_new(self):
        """ Make a deep copy of the individual, but with fitness set to None and assigned a new id. """
        return Individual(main_node=self.main_node.copy(), to_pipeline=self._to_pipeline)

    @classmethod
    def from_string(cls, string: str, primitive_set: dict, to_pipeline: Optional[Callable] = None):
        """ Construct an Individual from its `pipeline_str` representation.

        Parameters
        ----------
        string: str
            String formatted as `Individual.pipeline_str`.
        primitive_set: dict
            The dictionary defining all Terminals and Primitives.
        to_pipeline: Callable, optional (default=None)
            The function to convert the Individual into a pipeline representation.
            If `None`, the individuals `pipeline` property will not be available.

        Returns
        -------
        Individual
            An individual as defined by `str`.
        """
        expression = PrimitiveNode.from_string(string, primitive_set)
        return cls(expression, to_pipeline=to_pipeline)



"""
Contains mutation functions for genetic programming.
Each mutation function takes an individual and modifies it in-place.
"""
import random
from typing import Callable

from .components import Individual, DATA_TERMINAL
from .operations import random_primitive_node


def mut_replace_terminal(individual: Individual, primitive_set: dict) -> None:
    """ Mutates an Individual in-place by replacing one of its Terminals.

    Parameters
    ----------
    individual: Individual
        Individual to mutate in-place.
    primitive_set: dict
    """
    terminals = list([(i, t) for i, t in enumerate(individual.terminals) if len(primitive_set[t.identifier]) > 1])
    if len(terminals) == 0:
        raise ValueError("Individual has no terminals or no terminals suitable for mutation.")

    terminal_index, old_terminal = random.choice(terminals)
    acceptable_new_terminals = [t for t in primitive_set[old_terminal.identifier] if t.value != old_terminal.value]
    new_terminal = random.choice(acceptable_new_terminals)
    individual.replace_terminal(terminal_index, new_terminal)


def mut_replace_primitive(individual: Individual, primitive_set: dict) -> None:
    """ Mutates an Individual in-place by replacing one of its Primitives.

    Parameters
    ----------
    individual: Individual
        Individual to mutate in-place.
    primitive_set: dict
    """
    replaceable_primitives = [(i, p) for i, p in enumerate(individual.primitives)
                              if len(primitive_set[p._primitive.output]) > 1]
    if replaceable_primitives == 0:
        raise ValueError("Individual has no primitives which can be replaced with a different primitive.")

    primitive_index, old_primitive_node = random.choice(replaceable_primitives)
    primitive_node = random_primitive_node(output_type=old_primitive_node._primitive.output,
                                           primitive_set=primitive_set,
                                           exclude=old_primitive_node._primitive)
    individual.replace_primitive(primitive_index, primitive_node)


def mut_shrink(individual: Individual, primitive_set: dict = None) -> None:
    """ Mutates an Individual in-place by removing any number of primitive nodes, starting with the last.

    Parameters
    ----------
    individual: Individual
        Individual to mutate in-place.
    primitive_set: dict, optional
        Not used. Present to create a matching function signature with other mutations.
    """
    n_primitives = len(list(individual.primitives))
    if n_primitives == 1:
        raise ValueError("Can not shrink an individual with only one primitive.")

    n_primitives_to_cut = random.randint(1, n_primitives-1)
    current_primitive_node = individual.main_node
    primitives_left = n_primitives - 1
    while primitives_left > n_primitives_to_cut:
        current_primitive_node = current_primitive_node._data_node
        primitives_left -= 1
    current_primitive_node._data_node = DATA_TERMINAL


def mut_insert(individual: Individual, primitive_set: dict) -> None:
    """ Mutate an Individual in-place by inserting a PrimitiveNode at a random location, except as root node.

    Parameters
    ----------
    individual: Individual
        Individual to mutate in-place.
    primitive_set: dict
    """
    if len(list(individual.primitives)) == 1:
        parent_node = individual.main_node
    else:
        parent_node = random.choice(list(individual.primitives)[:-1])
    new_primitive_node = random_primitive_node(output_type=DATA_TERMINAL, primitive_set=primitive_set)
    new_primitive_node._data_node = parent_node._data_node
    parent_node._data_node = new_primitive_node


def random_valid_mutation_in_place(individual: Individual, primitive_set: dict) -> Callable:
    """ Apply a random valid mutation in place.

    The random mutation can be one of:

     - mut_random_primitive
     - mut_random_terminal, if the individual has at least one
     - mutShrink, if individual has at least two primitives
     - mutInsert

    Parameters
    ----------
    individual: Individual
      An individual to be mutated *in-place*.
    primitive_set: dict
      A dictionary defining the set of primitives and terminals.

    Returns
    -------
    Callable
        The mutation function used.
    """
    available_mutations = [mut_replace_primitive, mut_insert]
    if len(list(individual.primitives)) > 1:
        available_mutations.append(mut_shrink)
    if len([t for t in individual.terminals if len(primitive_set[t.identifier]) > 1]):
        available_mutations.append(mut_replace_terminal)

    mut_fn = random.choice(available_mutations)
    mut_fn(individual, primitive_set)

    return mut_fn



import logging
from functools import partial
from typing import Optional, Any, Tuple, Dict, List, Callable

import pandas as pd

from gama.genetic_programming.components import Individual
from gama.genetic_programming.operator_set import OperatorSet
from gama.logging.machine_logging import TOKENS, log_event
from gama.logging.utility_functions import MultiprocessingLogger
from gama.search_methods.base_search import BaseSearch
from gama.utilities.generic.async_evaluator import AsyncEvaluator

log = logging.getLogger(__name__)


class AsyncEA(BaseSearch):
    """ Perform asynchronous evolutionary optimization.

    Parameters
    ----------
    population_size: int, optional (default=50)
        Maximum number of individuals in the population at any time.

    max_n_evaluations: int, optional (default=None)
        If specified, only a maximum of `max_n_evaluations` individuals are evaluated.
        If None, the algorithm will be run until interrupted by the user or a timeout.

    restart_callback: Callable, optional (default=None)
        A function with signature () -> `bool` which returns `True` if search should be restarted.
    """

    def __init__(self,
                 population_size: Optional[int] = None,
                 max_n_evaluations: Optional[int] = None,
                 restart_callback: Optional[Callable] = None):
        super().__init__()
        # maps hyperparameter -> (set value, default)
        self._hyperparameters: Dict[str, Tuple[Any, Any]] = dict(
            population_size=(population_size, 50),
            restart_callback=(restart_callback, None),
            max_n_evaluations=(max_n_evaluations, None)
        )
        self.output = []

    def dynamic_defaults(self, x: pd.DataFrame, y: pd.DataFrame, time_limit: int):
        pass

    def search(self, operations: OperatorSet, start_candidates: List[Individual]):
        self.output = async_ea(operations, self.output, start_candidates, **self.hyperparameters)


def async_ea(
        operations: OperatorSet,
        output: List[Individual],
        start_candidates: List[Individual],
        restart_callback: Optional[Callable[[], bool]] = None,
        max_n_evaluations: Optional[int] = None,
        population_size: int = 50) -> List[Individual]:
    """ Perform asynchronous evolutionary optimization given the evolutionary operators in `operations`.

    Parameters
    ----------
    operations: OperatorSet
        An operator set with `evaluate`, `create`, `individual` and `eliminate` functions.
    output: List[Individual]
        A list which contains the set of best found individuals during search.
    start_candidates: List[Individual]
        A list with candidate individuals which should be used to start search from.
    restart_callback: Callable[[], bool], optional (default=None)
        A function with signature () -> bool which returns True if search should be restarted.
    max_n_evaluations: int, optional (default=None)
        If specified, only a maximum of `max_n_evaluations` individuals are evaluated.
        If None, the algorithm will be run indefinitely.
    population_size: int (default=50)
        Maximum number of individuals in the population at any time.

    Returns
    -------
    List[Individual]
        The individuals currently in the population.
    """
    if max_n_evaluations is not None and max_n_evaluations <= 0:
        raise ValueError("'n_evaluations' must be non-negative or None, but was {}.".format(max_n_evaluations))

    max_population_size = population_size
    logger = MultiprocessingLogger()

    evaluate_log = partial(operations.evaluate, logger=logger)

    current_population = output
    n_evaluated_individuals = 0

    with AsyncEvaluator() as async_:
        should_restart = True
        while should_restart:
            should_restart = False
            current_population[:] = []
            log.info('Starting EA with new population.')
            for individual in start_candidates:
                async_.submit(evaluate_log, individual)

            while (max_n_evaluations is None) or (n_evaluated_individuals < max_n_evaluations):
                future = operations.wait_next(async_)
                logger.flush_to_log(log)
                if future.exception is None:
                    individual = future.result
                    current_population.append(individual)
                    if len(current_population) > max_population_size:
                        to_remove = operations.eliminate(current_population, 1)
                        log_event(log, TOKENS.EA_REMOVE_IND, to_remove[0])
                        current_population.remove(to_remove[0])

                if len(current_population) > 1:
                    new_individual = operations.create(current_population, 1)[0]
                    async_.submit(evaluate_log, new_individual)

                should_restart = (restart_callback is not None and restart_callback())
                n_evaluated_individuals += 1
                if should_restart:
                    log.info("Restart criterion met. Restarting with new random population.")
                    log_event(log, TOKENS.EA_RESTART, n_evaluated_individuals)
                    start_candidates = [operations.individual() for _ in range(max_population_size)]
                    break


    return current_population



from collections import defaultdict
from datetime import datetime
from typing import List, Optional, Tuple, Dict, Callable

import pandas as pd

from gama.configuration.classification import clf_config
from gama.configuration.parser import pset_from_config, merge_configurations
from gama.configuration.regression import reg_config
from gama.genetic_programming.components import Individual
from gama.logging.machine_logging import PLE_START, PLE_DELIM, PLE_END, TOKENS, METHOD_TOKENS
from gama.logging import TIME_FORMAT


pset, _ = pset_from_config(merge_configurations(clf_config, reg_config))


class GamaReport:
    """ Contains information over an information trace as parsed from a GAMA analysis log. """

    def __init__(
            self,
            logfile: Optional[str] = None,
            log_lines: Optional[List[str]] = None,
            name: Optional[str] = None
    ):
        """ Parses the logfile or log lines provided. Must provide exactly one of 'logfile' or 'loglines'.

        Parameters
        ----------
        logfile: str, optional (default=None)
            Path to the log file. If not specified, loglines must be provided.
        log_lines: List[str], optional (default=None)
            A list with each element one line from the log file. If not specified, logfile must be provided.
        name: str, optional (default=None)
            Name of the report. If set to None, defaults to `logfile` if it is not None else 'nameless'.
        """
        if logfile is None and log_lines is None:
            raise ValueError("Either 'logfile' or 'loglines' must be provided. Both are None.")
        if logfile is not None and log_lines is not None:
            raise ValueError("Exactly one of 'logfile' and 'loglines' may be provided at once.")

        if logfile is not None:
            log_lines = _find_new_lines(logfile)

        self._lines_read = len(log_lines)
        self._individuals = None
        self.name = name if name is not None else (logfile if logfile is not None else 'nameless')

        events_by_type = _lines_to_dict(log_lines)

        if len(events_by_type[TOKENS.INIT]) == 0:
            raise ValueError("The log must contain at least contain an INIT string.")

        self.metrics, self.search_method, self.postprocessing, self._filename \
            = _find_metric_configuration(events_by_type[TOKENS.INIT])
        self.phases: List[Tuple[str, str, datetime, float]] = _find_phase_information(events_by_type)
        search_start = self.phases[1][2] if len(self.phases) > 1 else None
        self.evaluations: pd.DataFrame = _evaluations_to_dataframe(events_by_type[TOKENS.EVALUATION_RESULT],
                                                                   metric_names=self.metrics,
                                                                   search_start=search_start)

        # This can take a while for long logs (e.g. ~1sec for 10k individuals)
        self.individuals: Dict[str, Individual] = {
            id_: Individual.from_string(pipeline, pset)
            for id_, pipeline in zip(self.evaluations.id, self.evaluations.pipeline)
        }

        parse_method_data: Dict[str, Callable[..., pd.DataFrame]] = defaultdict(
            lambda: lambda *args: None,
            AsynchronousSuccessiveHalving=_ASHA_data_to_dataframe
        )
        # search_method is formatted like NAME(kwargs) where kwargs could contain additional parentheses.
        method_name, _ = self.search_method.split('(', maxsplit=1)
        method_token = METHOD_TOKENS.get(method_name)
        self.method_data = parse_method_data[method_name](events_by_type[method_token], self.metrics)

        self.incomplete = (len(self.phases) < 3)

    def update(self) -> bool:
        new_lines = _find_new_lines(self._filename, start_from=self._lines_read)
        if len(new_lines) > 0:
            self._lines_read += len(new_lines)
            print(f'read {len(new_lines)} new lines')
            events_by_type = _lines_to_dict(new_lines)
            search_start = 0 if len(self.evaluations) == 0 else self.evaluations.start.min()
            new_evaluations = _evaluations_to_dataframe(
                events_by_type[TOKENS.EVALUATION_RESULT],
                metric_names=self.metrics,
                search_start=search_start,
                start_n=self.evaluations.n.max() + 1
            )
            self.evaluations = pd.concat([self.evaluations, new_evaluations])
            for metric in self.metrics:
                self.evaluations[f'{metric}_cummax'] = self.evaluations[metric].cummax()
            self.individuals.update({
                id_: Individual.from_string(pipeline, pset)
                for id_, pipeline in zip(new_evaluations.id, new_evaluations.pipeline)
            })
        return len(new_lines) > 0


def _lines_to_dict(log_lines: List[str]):
    # Find the Parseable Log Events and discard their start/end tokens.
    ple_lines = [line.split(PLE_DELIM)[1:-1] for line in log_lines
                 if line.startswith(PLE_START) and line.endswith(f'{PLE_END}')]

    events_by_type = defaultdict(list)
    for token, *event in ple_lines:
        events_by_type[token].append(event)

    return events_by_type


def _find_new_lines(logfile: str, start_from: int = 0):
    with open(logfile, 'r') as fh:
        log_lines = [line.rstrip() for line in fh.readlines()]
    new_lines = log_lines[start_from:]
    return new_lines


def _find_metric_configuration(init_lines: List[List[str]]) -> Tuple[List[str], str, str, str]:
    scoring, regularize_length, *_, filename, _, search, postprocessing = init_lines[0][0].split(',')
    _, metric = scoring.split('=')
    _, regularize = regularize_length.split('=')
    _, search = search.split('=', maxsplit=1)
    _, filename = filename.split('=', maxsplit=1)
    _, postprocessing = postprocessing.split('=', maxsplit=1)
    if bool(regularize):
        return [metric, 'length'], search, postprocessing, filename
    else:
        return [metric], search, postprocessing, filename


def _find_phase_information(events_by_type: Dict[str, List[str]]) -> List[Tuple[str, str, datetime, float]]:
    """ For each phase (e.g. search), find the type used (e.g. ASHA) and its duration. """
    phases = ['preprocessing', 'search', 'postprocess']
    phase_info = []
    # Events as phase;algorithm;logtime
    for phase in phases:
        start_phase_events = [event for event in events_by_type[TOKENS.PHASE_START] if phase in event]
        end_phase_events = [event for event in events_by_type[TOKENS.PHASE_END] if phase in event]
        if start_phase_events == [] or end_phase_events == []:
            # the phase has either not yet started, or not yet completed. Then this is also true for later phases.
            break
        start_phase, end_phase = start_phase_events[0], end_phase_events[0]
        _, _, start_time = start_phase
        _, algorithm, end_time = end_phase
        duration = (datetime.strptime(end_time, TIME_FORMAT) - datetime.strptime(start_time, TIME_FORMAT))
        phase_info.append([phase, algorithm, datetime.strptime(start_time, TIME_FORMAT), duration.total_seconds()])
    return phase_info


def _evaluations_to_dataframe(evaluation_lines: List[List[str]],
                              metric_names: Optional[List[str]] = None,
                              search_start: datetime = None,
                              start_n: int = 0) -> pd.DataFrame:
    """ Create a dataframe with all pipeline evaluations as parsed from EVAL events in the log. """
    evaluations = []
    for i, line in enumerate(evaluation_lines, start=start_n):
        time, duration, process_duration, fitness, id_, pipeline_str, log_time = line
        # Fitness logged as '(metric1, metric2, ..., metriclast)'
        metrics_values = [float(value) for value in fitness[1:-1].split(',')]
        evaluations.append([i, time, float(duration), *metrics_values, pipeline_str, id_])

    if metric_names is None:
        metric_names = [f'metric_{m_i}' for m_i in range(len(metrics_values))]
    column_names = ['n', 'start', 'duration', *metric_names, 'pipeline', 'id']
    df = pd.DataFrame(evaluations, columns=column_names)
    for metric in metric_names:
        df[f'{metric}_cummax'] = df[metric].cummax()

    df.start = pd.to_datetime(df.start)
    df.duration = pd.to_timedelta(df.duration, unit='s')
    search_start = search_start if search_start is not None else df.start.min()
    df['relative_end'] = ((df.start + df.duration) - search_start).dt.total_seconds()
    return df


def _ASHA_data_to_dataframe(asha_lines: List[List[str]],
                            metric_names: Optional[List[str]] = None) -> pd.DataFrame:
    asha_data = []
    for i, line in enumerate(asha_lines):
        rung, duration, fitness, id_, pipeline_str, log_time = line
        metrics_values = [float(value) for value in fitness[1:-1].split(',')]
        asha_data.append([i, rung, float(duration), *metrics_values, pipeline_str, id_])

    if metric_names is None:
        metric_names = [f'metric_{m_i}' for m_i in range(len(metrics_values))]
    column_names = ['n', 'rung', 'duration', *metric_names, 'pipeline', 'id']
    df = pd.DataFrame(asha_data, columns=column_names)
    df.duration = pd.to_timedelta(df.duration, unit='s')

    return df



import time

import pytest
import pandas as pd
from sklearn.datasets import load_iris
from sklearn.tree import DecisionTreeClassifier
from sklearn.metrics import accuracy_score, log_loss
from sklearn.preprocessing import OneHotEncoder

from gama.utilities.metrics import scoring_to_metric
from gama.genetic_programming.compilers.scikitlearn import \
    cross_val_predict_score, evaluate_individual, compile_individual, evaluate_pipeline
from tests.unit.unit_fixtures import BernoulliNBStandardScaler, pset


def test_cross_val_predict_score():
    estimator = DecisionTreeClassifier()
    x, y = load_iris(return_X_y=True)
    y_ohe = OneHotEncoder().fit_transform(y.reshape(-1, 1))
    x, y = pd.DataFrame(x), pd.Series(y)

    metrics = scoring_to_metric(['accuracy', 'log_loss'])
    predictions, scores = cross_val_predict_score(estimator, x, y, metrics=metrics)
    accuracy, logloss = scores

    assert accuracy_score(y_ohe, predictions) == pytest.approx(accuracy)
    assert -1 * log_loss(y_ohe, predictions) == pytest.approx(logloss)


def test_evaluate_individual(BernoulliNBStandardScaler, mocker):
    import datetime
    reported_start_time = datetime.datetime.now()
    def fake_evaluate(*args, **kwargs):
        # (scores), start, walltime, processtime
        return (1.0, -0.5), reported_start_time, 0.5, 0.7

    mocker.patch('gama.genetic_programming.compilers.scikitlearn.evaluate_pipeline', new=fake_evaluate)
    individual = evaluate_individual(BernoulliNBStandardScaler, evaluate_pipeline_length=True)
    assert individual == BernoulliNBStandardScaler
    assert hasattr(individual, 'fitness')
    assert individual.fitness.values == (1.0, -0.5, -2)
    assert individual.fitness.start_time == reported_start_time
    assert individual.fitness.wallclock_time == 0.5
    assert individual.fitness.process_time == 0.7


def test_compile_individual(BernoulliNBStandardScaler):
    from sklearn.naive_bayes import BernoulliNB
    from sklearn.preprocessing import StandardScaler, MinMaxScaler

    pipeline = compile_individual(BernoulliNBStandardScaler)
    assert 2 == len(pipeline.steps)
    assert isinstance(pipeline.steps[0][1], StandardScaler)
    assert isinstance(pipeline.steps[1][1], BernoulliNB)

    extended_pipeline = compile_individual(BernoulliNBStandardScaler, preprocessing_steps=[MinMaxScaler()])
    assert 3 == len(extended_pipeline.steps)
    assert isinstance(extended_pipeline.steps[0][1], MinMaxScaler)
    assert isinstance(extended_pipeline.steps[1][1], StandardScaler)
    assert isinstance(extended_pipeline.steps[2][1], BernoulliNB)


def test_evaluate_pipeline(BernoulliNBStandardScaler):
    x, y = load_iris(return_X_y=True)
    x, y = pd.DataFrame(x), pd.Series(y)

    scores, start, wallclock, process = evaluate_pipeline(
        BernoulliNBStandardScaler.pipeline, x, y, timeout=60, deadline=time.time()+60,
        metrics=scoring_to_metric('accuracy'))
    assert 1 == len(scores)



from typing import NamedTuple

DATA_TERMINAL = 'data'

# Specifies a specific value for a specific type or input, e.g. a value for a hyperparameter for an algorithm.
Terminal = NamedTuple("Terminal",
                      [("value", object),
                       ("output", str),
                       ("identifier", str)])


def str_format_terminal_value(terminal) -> str:
    if isinstance(terminal.value, str):
        return "'{}'".format(terminal.value)
    elif callable(terminal.value):
        return "{}".format(terminal.value.__name__)
    else:
        return str(terminal.value)


def terminal__str__(terminal) -> str:
    """ str: e.g. "tol=0.5" """
    return "{}={}".format(terminal.output, str_format_terminal_value(terminal))


def terminal__repr__(terminal) -> str:
    """ str: e.g. "FastICA.tol=0.5". If the hyperparameter is shared across primitives, there is no prefix. """
    return "{}={}".format(terminal.identifier, str_format_terminal_value(terminal))


Terminal.__str__ = terminal__str__
Terminal.__repr__ = terminal__repr__



import pytest

from gama.postprocessing import NoPostProcessing


def test_no_post_processing():
    """ Test that NoPostProcessing does nothing and no model is returned. """
    postprocessing = NoPostProcessing()
    postprocessing.dynamic_defaults(None)
    model = postprocessing.post_process()
    assert pytest.approx(0.0) == postprocessing.time_fraction
    assert model is None



import numpy as np

from sklearn.naive_bayes import GaussianNB, BernoulliNB, MultinomialNB
from sklearn.tree import DecisionTreeClassifier
from sklearn.ensemble import ExtraTreesClassifier, RandomForestClassifier, GradientBoostingClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import LinearSVC
from sklearn.linear_model import LogisticRegression
from sklearn.cluster import FeatureAgglomeration
from sklearn.preprocessing import MaxAbsScaler, MinMaxScaler, Normalizer, PolynomialFeatures, RobustScaler, \
    StandardScaler, Binarizer
from sklearn.kernel_approximation import Nystroem, RBFSampler
from sklearn.decomposition import PCA, FastICA
from sklearn.feature_selection import SelectFwe, SelectPercentile, f_classif, VarianceThreshold

# This selection of operators and hyperparameters is currently most of what TPOT supports, for comparison.

clf_config = {

    'alpha': [1e-3, 1e-2, 1e-1, 1., 10., 100.],
    'fit_prior': [True, False],
    'min_samples_split': range(2, 21),
    'min_samples_leaf': range(1, 21),

    # Classifiers
    GaussianNB: {
    },

    BernoulliNB: {
        'alpha': [],
        'fit_prior': []
    },

    MultinomialNB: {
        'alpha': [],
        'fit_prior': []
    },

    DecisionTreeClassifier: {
        'criterion': ["gini", "entropy"],
        'max_depth': range(1, 11),
        'min_samples_split': [],
        'min_samples_leaf': []
    },

    ExtraTreesClassifier: {
        'n_estimators': [100],
        'criterion': ["gini", "entropy"],
        'max_features': [0.05, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
        'min_samples_split': [],
        'min_samples_leaf': [],
        'bootstrap': [True, False]
    },

    RandomForestClassifier: {
        'n_estimators': [100],
        'criterion': ["gini", "entropy"],
        'max_features': [0.05, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
        'min_samples_split': range(2, 21),
        'min_samples_leaf': range(1, 21),
        'bootstrap': [True, False]
    },

    GradientBoostingClassifier: {
        'n_estimators': [100],
        'learning_rate': [1e-3, 1e-2, 1e-1, 0.5, 1.],
        'max_depth': range(1, 11),
        'min_samples_split': range(2, 21),
        'min_samples_leaf': range(1, 21),
        'subsample': [0.05, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
        'max_features': np.arange(0.05, 1.01, 0.05)
    },

    KNeighborsClassifier: {
        'n_neighbors': range(1, 51),
        'weights': ["uniform", "distance"],
        'p': [1, 2]
    },

    LinearSVC: {
        'penalty': ["l1", "l2"],
        'loss': ["hinge", "squared_hinge"],
        'dual': [False, True],
        'tol': [1e-5, 1e-4, 1e-3, 1e-2, 1e-1],
        'C': [1e-4, 1e-3, 1e-2, 1e-1, 0.5, 1., 5., 10., 15., 20., 25.],
        'param_check': [lambda params: (not params['dual'] or params['penalty'] == "l2")
                                       and not (params['penalty'] == "l1" and params['loss'] == "hinge")
                                       and not (
                    params['penalty'] == "l2" and params['loss'] == "hinge" and not params['dual'])]
    },

    LogisticRegression: {
        'penalty': ["l1", "l2"],
        'C': [1e-4, 1e-3, 1e-2, 1e-1, 0.5, 1., 5., 10., 15., 20., 25.],
        'dual': [False, True],
        'param_check': [lambda params: not params['dual'] or params['penalty'] == "l2"]
    },

    # Preprocesssors
    Binarizer: {
        'threshold': np.arange(0.0, 1.01, 0.05)
    },

    FastICA: {
        'tol': np.arange(0.0, 1.01, 0.05)
    },

    FeatureAgglomeration: {
        'linkage': ['ward', 'complete', 'average'],
        'affinity': ['euclidean', 'l1', 'l2', 'manhattan', 'cosine', 'precomputed'],
        'param_check': [lambda params: (not params['linkage'] == "ward") or params['affinity'] == "euclidean"]
    },

    MaxAbsScaler: {
    },

    MinMaxScaler: {
    },

    Normalizer: {
        'norm': ['l1', 'l2', 'max']
    },

    Nystroem: {
        'kernel': ['rbf', 'cosine', 'chi2', 'laplacian', 'polynomial', 'poly', 'linear', 'additive_chi2', 'sigmoid'],
        'gamma': np.arange(0.0, 1.01, 0.05),
        'n_components': range(1, 11)
    },

    PCA: {
        'svd_solver': ['randomized'],
        'iterated_power': range(1, 11)
    },

    PolynomialFeatures: {
        'degree': [2],
        'include_bias': [False],
        'interaction_only': [False]
    },

    RBFSampler: {
        'gamma': np.arange(0.0, 1.01, 0.05)
    },

    RobustScaler: {
    },

    StandardScaler: {
    },

    # Selectors
    SelectFwe: {
        'alpha': np.arange(0, 0.05, 0.001),
        'score_func': {
            f_classif: None
        }
    },

    SelectPercentile: {
        'percentile': range(1, 100),
        'score_func': {
            f_classif: None
        }
    },

    VarianceThreshold: {
        'threshold': np.arange(0.05, 1.01, 0.05)
    }
}



from gama.utilities.generic.paretofront import ParetoFront


def test_pareto_initialization_empty():
    """ Test initialization of empty front. """
    pf = ParetoFront()

    assert list(pf) == []
    assert str(pf) == '[]'
    assert repr(pf) == 'ParetoFront([])'


def test_pareto_initialization_pareto_front():
    """ Test initialization with a list which only has Pareto front elements. """
    list_ = [(1, 2, 3), (3, 2, 1), (0, 5, 0)]
    pf = ParetoFront(list_)

    assert list(pf) == [(1, 2, 3), (3, 2, 1), (0, 5, 0)]
    assert str(pf) == '[(1, 2, 3), (3, 2, 1), (0, 5, 0)]'
    assert repr(pf) == 'ParetoFront([(1, 2, 3), (3, 2, 1), (0, 5, 0)])'


def test_pareto_initialization_with_inferiors():
    """" Test initialization with a list containing elements that should not be in the Pareto front. """
    list_ = [(1, 2), (4, 3), (4, 5), (5, 4)]
    pf = ParetoFront(list_)

    assert list(pf) == [(4, 5), (5, 4)]
    assert str(pf) == '[(4, 5), (5, 4)]'
    assert repr(pf) == 'ParetoFront([(4, 5), (5, 4)])'


def test_pareto_initialization_with_duplicates():
    """ Test initialization with duplicate elements in list. """
    list_ = [(1, 2), (3, 1), (1, 2)]
    pf = ParetoFront(list_)

    assert list(pf) == [(1, 2), (3, 1)]
    assert str(pf) == '[(1, 2), (3, 1)]'
    assert repr(pf) == 'ParetoFront([(1, 2), (3, 1)])'


def test_pareto_update_unique():
    """ Test creating Pareto front by updating one by one. """
    list_ = [(1, 2, 3), (3, 2, 1), (0, 5, 0)]
    pf = ParetoFront()

    for i in range(len(list_)):
        pf.update(list_[i])
        assert list(pf) == list_[:i+1]


def test_pareto_front_clear():
    """ Tests that the calling clear empties the Pareto front. """
    pf = ParetoFront([(1, 2), (2, 1)])
    assert list(pf) == [(1, 2), (2, 1)]

    pf.clear()
    assert list(pf) == []


def test_pareto_front_custom_function():
    """ Test construction of Pareto front with custom object and value function. """
    class A:
        def __init__(self, v1, v2):
            self.v1 = v1
            self.v2 = v2

    item1, item2, item3 = A(1, 2), A(2, 1), A(3, 1)
    pf = ParetoFront(get_values_fn=lambda x: (x.v1, x.v2))

    pf.update(item1)
    assert list(pf) == [item1]

    pf.update(item2)
    assert list(pf) == [item1, item2]

    pf.update(item3)
    assert list(pf) == [item1, item3]




from abc import ABC
from typing import List, Union, Dict, Any

import pandas as pd


from gama.genetic_programming.components import Individual


class BasePostProcessing(ABC):
    """ All post-processing methods should be derived from this class.
    This class should not be directly used to configure GAMA.
    """

    def __init__(self, time_fraction: float):
        """

        Parameters
        ----------
        time_fraction: float
            The fraction of total time that should be reserved for this post-processing step.
        """
        self.time_fraction: float = time_fraction
        self._hyperparameters = {}

    def __str__(self):
        # Not sure if I should report actual used hyperparameters (i.e. include default), or only those set by user.
        user_set_hps = {parameter: set_value
                        for parameter, (set_value, default) in self._hyperparameters.items()
                        if set_value is not None}
        hp_configuration = ','.join([f"{name}={value}" for (name, value) in user_set_hps.items()])
        return f"{self.__class__.__name__}({hp_configuration})"

    @property
    def hyperparameters(self) -> Dict[str, Any]:
        """ Hyperparameter (name, value) pairs value determined by user > dynamic default > static default.

         Dynamic default values only considered if `dynamic_defaults` has been called.
         """
        return {parameter: set_value if set_value is not None else default
                for parameter, (set_value, default) in self._hyperparameters.items()}

    def _overwrite_hyperparameter_default(self, hyperparameter: str, value: Any):
        set_value, default_value = self._hyperparameters[hyperparameter]
        self._hyperparameters[hyperparameter] = (set_value, value)

    def dynamic_defaults(self, gama: 'Gama'):
        pass

    def post_process(
            self,
            x: pd.DataFrame,
            y: Union[pd.DataFrame, pd.Series],
            timeout: float,
            selection: List[Individual]) -> 'model':
        """
        Parameters
        ----------
        x: pd.DataFrame
            all training features
        y: Union[pd.DataFrame, pd.Series]
            all training labels
        timeout: float
            allowed time in seconds for post-processing
        selection: List[Individual]
            individuals selected by the search space, ordered best first

        Returns
        -------
        Any
            A model with `predict` and optionally `predict_proba`.
        """
        raise NotImplementedError("Method must be implemented by child class.")



""" Selection operators. """
import random
import numpy

from gama.utilities.generic.paretofront import ParetoFront
from gama.genetic_programming.crossover import _valid_crossover_functions


def create_from_population(operator_shell, pop, n, cxpb, mutpb):
    """ Creates n new individuals based on the population. Can apply both crossover and mutation. """
    offspring = []
    for _ in range(n):
        ind1, ind2 = random.sample(pop, k=2)
        if random.random() < cxpb and len(_valid_crossover_functions(ind1, ind2)) > 0:
            ind1 = operator_shell.mate(ind1, ind2)
        else:
            ind1 = operator_shell.mutate(ind1)
        offspring.append(ind1)
    return offspring


def select_from_pareto(population, select_n, pareto_fronts_n):
    pareto_fronts = []
    population_left = population
    while len(pareto_fronts) != pareto_fronts_n:
        pareto_fronts.append(ParetoFront(start_list=population_left, get_values_fn=lambda ind: ind.fitness.values))
        population_left = [ind for ind in population_left if ind not in pareto_fronts[-1]]
        if len(population_left) == 0 and len(pareto_fronts) != pareto_fronts_n:
            # log.debug("Desired amount of pareto fronts could not be constructed.")
            break

    selected_individuals = []
    for _ in range(select_n):
        if len(pareto_fronts) == 1:
            selected_front = pareto_fronts[0]
        elif len(pareto_fronts) == 2:
            selected_front = numpy.random.choice(pareto_fronts, p=[2/3, 1/3])
        elif len(pareto_fronts) > 2:
            n = len(pareto_fronts)
            selected_front = numpy.random.choice(pareto_fronts, p=[1/2**i for i in range(1, n)] + [1/2**(n-1)])
        else:
            raise RuntimeError("Did not create any pareto front.")

        selected_individuals.append(random.choice(selected_front))
    return selected_individuals


def eliminate_from_pareto(pop, n):
    # For now we only eliminate one at a time so this will do.
    if n != 1:
        raise NotImplemented("Currently only n=1 is supported.")

    def inverse_fitness(ind):
        return [-value for value in ind.fitness.values]

    pareto_worst = ParetoFront(pop, inverse_fitness)
    return [random.choice(pareto_worst)]



from collections.abc import Sequence
import logging

from gama.logging.machine_logging import TOKENS, log_event
from gama.utilities.generic.async_evaluator import AsyncEvaluator
from .components import Individual

log = logging.getLogger(__name__)


class OperatorSet:
    """ Provides a thin layer for ea operators for logging, callbacks and safety. """

    def __init__(self, mutate, mate, create_from_population, create_new, compile_, eliminate, evaluate_callback, max_retry=50):
        """

        :param mutate:
        :param mate:
        :param create:
        :param create_new:
        """

        self._mutate = mutate
        self._mate = mate
        self._create_from_population = create_from_population
        self._create_new = create_new
        self._compile = compile_
        self._safe_compile = None
        self._eliminate = eliminate
        self._max_retry = max_retry
        self._evaluate = None
        self._evaluate_callback = evaluate_callback

        self._seen_individuals = {}

    def wait_next(self, async_evaluator):
        future = async_evaluator.wait_next()
        if future.result is not None:
            result = future.result
            if not isinstance(result, Sequence):
                individual = result
                log_event(log, TOKENS.EVALUATION_RESULT, individual.fitness.start_time,
                          individual.fitness.wallclock_time, individual.fitness.process_time,
                          individual.fitness.values, individual._id, individual.pipeline_str())
            else:
                # For now, we leave logging for non-standard results to the caller (e.g. rungs in ASHA)
                individual = result[0]
            if self._evaluate_callback is not None:
                self._evaluate_callback(individual)
        elif future.exception is not None:
            log.warning(f'Encountered exception while evaluation individual: {str(future.exception)}.')
        return future

    def try_until_new(self, operator, *args, **kwargs):
        for _ in range(self._max_retry):
            individual, log_args = operator(*args, **kwargs)
            if str(individual.main_node) not in self._seen_individuals:
                #self._seen_individuals[str(individual.main_node)] = individual
                return individual, log_args
        else:
            log.debug("Could not create a new individual from 50 iterations of {}".format(operator.__name__))
            # For progress on solving this, see #11
            return individual, log_args

    def mate(self, individual1: Individual, individual2: Individual, *args, **kwargs):
        def mate_with_log():
            new_individual1, new_individual2 = individual1.copy_as_new(), individual2.copy_as_new()
            self._mate(new_individual1, new_individual2, *args, **kwargs)
            log_args = [TOKENS.CROSSOVER, new_individual1._id, individual1._id, individual2._id]
            return new_individual1, log_args

        individual, log_args = self.try_until_new(mate_with_log)
        log_event(log, *log_args)
        return individual

    def mutate(self, individual: Individual, *args, **kwargs):
        def mutate_with_log():
            new_individual = individual.copy_as_new()
            mutator = self._mutate(new_individual, *args, **kwargs)
            log_args = [TOKENS.MUTATION, new_individual._id, individual._id, mutator.__name__]
            return new_individual, log_args

        individual, log_args = self.try_until_new(mutate_with_log)
        log_event(log, *log_args)
        return individual

    def individual(self, *args, **kwargs):
        expression = self._create_new(*args, **kwargs)
        compile_ = self._safe_compile if self._safe_compile is not None else self._compile
        return Individual(expression, to_pipeline=compile_)

    def create(self, *args, **kwargs):
        return self._create_from_population(self, *args, **kwargs)

    def eliminate(self, *args, **kwargs):
        return self._eliminate(*args, **kwargs)



""" Contains full system tests for GamaClassifier """
import numpy as np
import pandas as pd
import pytest
from typing import Type

from sklearn.datasets import load_wine, load_breast_cancer
from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score, log_loss

from gama.postprocessing import EnsemblePostProcessing
from gama.search_methods import asha, async_ea
from gama.search_methods.asha import AsynchronousSuccessiveHalving
from gama.search_methods.async_ea import AsyncEA
from gama.search_methods.base_search import BaseSearch
from gama.search_methods.random_search import RandomSearch
from gama.utilities.generic.stopwatch import Stopwatch
from gama import GamaClassifier

import warnings
warnings.filterwarnings("error")

FIT_TIME_MARGIN = 1.1

# While we could derive statistics dynamically, we want to know if any changes ever happen, so we save them statically.
breast_cancer = dict(
    name='breast_cancer',
    load=load_breast_cancer,
    test_size=143,
    n_classes=2,
    base_accuracy=0.62937,
    base_log_loss=12.80138
)

breast_cancer_missing = dict(
    name='breast_cancer_missing',
    load=load_breast_cancer,
    test_size=143,
    n_classes=2,
    base_accuracy=0.62937,
    base_log_loss=12.80138
)

wine = dict(
    name='wine',
    load=load_wine,
    test_size=45,
    n_classes=3,
    base_accuracy=0.4,
    base_log_loss=20.72326,
)

iris_arff = dict(
    name='iris',
    train='tests/data/iris_train.arff',
    test='tests/data/iris_test.arff',
    test_size=50,
    n_classes=3,
    base_accuracy=0.3333,
    base_log_loss=1.09861
)

diabetes_arff = dict(
    name='diabetes',
    train='tests/data/diabetes_train.arff',
    test='tests/data/diabetes_test.arff',
    test_size=150,
    n_classes=2,
    base_accuracy=0.65104,
    base_log_loss=0.63705
)


def _test_dataset_problem(
        data,
        metric: str,
        arff: bool = False,
        y_type: Type = pd.DataFrame,
        search: BaseSearch = AsyncEA(),
        missing_values: bool = False):
    """

    :param data:
    :param metric:
    :param arff:
    :param y_type: pd.DataFrame, pd.Series, np.ndarray or str
    :return:
    """
    gama = GamaClassifier(random_state=0, max_total_time=60, scoring=metric, search_method=search, n_jobs=1,
                          cache_dir=f"{data['name']}_arff_{arff}_m_{metric}_y_{y_type.__name__}_s_{search.__class__.__name__}",
                          post_processing_method=EnsemblePostProcessing(ensemble_size=5))
    if arff:
        train_path = 'tests/data/{}_train.arff'.format(data['name'])
        test_path = 'tests/data/{}_test.arff'.format(data['name'])

        X, y = data['load'](return_X_y=True)
        X_train, X_test, y_train, y_test = train_test_split(X, y, stratify=y, random_state=0)
        y_test = [str(val) for val in y_test]

        with Stopwatch() as sw:
            gama.fit_arff(train_path)
        class_predictions = gama.predict_arff(test_path)
        class_probabilities = gama.predict_proba_arff(test_path)
        gama_score = gama.score_arff(test_path)
    else:
        X, y = data['load'](return_X_y=True)
        if y_type == str:
            databunch = data['load']()
            y = np.asarray([databunch.target_names[c_i] for c_i in databunch.target])
        if y_type in [pd.Series, pd.DataFrame]:
            y = y_type(y)

        X_train, X_test, y_train, y_test = train_test_split(X, y, stratify=y, random_state=0)
        if missing_values:
            X_train[1:300:2, 0] = X_train[2:300:5, 1] = float("NaN")
            X_test[1:100:2, 0] = X_test[2:100:5, 1] = float("NaN")

        with Stopwatch() as sw:
            gama.fit(X_train, y_train)
        class_predictions = gama.predict(X_test)
        class_probabilities = gama.predict_proba(X_test)
        gama_score = gama.score(X_test, y_test)

    assert 60 * FIT_TIME_MARGIN > sw.elapsed_time, 'fit must stay within 110% of allotted time.'

    assert isinstance(class_predictions, np.ndarray), 'predictions should be numpy arrays.'
    assert (data['test_size'],) == class_predictions.shape, 'predict should return (N,) shaped array.'

    accuracy = accuracy_score(y_test, class_predictions)
    # Majority classifier on this split achieves 0.6293706293706294
    print(data['name'], metric, 'accuracy:', accuracy)
    assert data['base_accuracy'] <= accuracy, 'predictions should be at least as good as majority class.'

    assert isinstance(class_probabilities, np.ndarray), 'probability predictions should be numpy arrays.'
    assert (data['test_size'], data['n_classes']) == class_probabilities.shape, ('predict_proba should return'
                                                                                 ' (N,K) shaped array.')

    # Majority classifier on this split achieves 12.80138131184662
    logloss = log_loss(y_test, class_probabilities)
    print(data['name'], metric, 'log-loss:', logloss)
    assert data['base_log_loss'] >= logloss, 'predictions should be at least as good as majority class.'

    score_to_match = logloss if metric == 'log_loss' else accuracy
    assert score_to_match == pytest.approx(gama_score)
    gama.delete_cache()


def test_binary_classification_accuracy():
    """ GamaClassifier can do binary classification with predict metric from numpy data. """
    _test_dataset_problem(breast_cancer, 'accuracy')


def test_binary_classification_accuracy_asha():
    """ GamaClassifier can do binary classification with predict metric from numpy data using ASHA search. """
    _test_dataset_problem(breast_cancer, 'accuracy', search=AsynchronousSuccessiveHalving())


def test_binary_classification_accuracy_random_search():
    """ GamaClassifier can do binary classification with predict metric from numpy data using ASHA search. """
    _test_dataset_problem(breast_cancer, 'accuracy', search=RandomSearch())


def test_binary_classification_logloss():
    """ GamaClassifier can do binary classification with predict-proba metric from numpy data. """
    _test_dataset_problem(breast_cancer, 'log_loss')


def test_multiclass_classification_accuracy():
    """ GamaClassifier can do multi-class with predict metric from numpy data. """
    _test_dataset_problem(wine, 'accuracy')


def test_multiclass_classification_logloss():
    """ GamaClassifier can do multi-class with predict-proba metric from numpy data. """
    _test_dataset_problem(wine, 'log_loss')


def test_string_label_classification_accuracy():
    """ GamaClassifier can work with string-like target labels when using predict-metric from numpy data. """
    _test_dataset_problem(breast_cancer, 'accuracy', y_type=str)


def test_string_label_classification_log_loss():
    """ GamaClassifier can work with string-type target labels when using predict-proba metric from numpy data. """
    _test_dataset_problem(breast_cancer, 'log_loss', y_type=str)


def test_missing_value_classification_arff():
    """ GamaClassifier handles missing data. """
    _test_dataset_problem(breast_cancer_missing, 'log_loss', arff=True)


def test_missing_value_classification():
    """ GamaClassifier handles missing data from numpy data. """
    _test_dataset_problem(breast_cancer_missing, 'log_loss', missing_values=True)



from typing import List

import pandas as pd

from gama.genetic_programming.components import Individual
from gama.postprocessing.base_post_processing import BasePostProcessing


class BestFitPostProcessing(BasePostProcessing):
    """ Post processing technique which trains the best found single pipeline. """

    def __init__(self, time_fraction: float = 0.1):
        super().__init__(time_fraction)

    def post_process(self, x: pd.DataFrame, y: pd.Series, timeout: float, selection: List[Individual]) -> 'model':
        return selection[0].pipeline.fit(x, y)



import dash

external_stylesheets = ['https://codepen.io/chriddyp/pen/bWLwgP.css']

dash_app = dash.Dash(__name__, external_stylesheets=external_stylesheets)
dash_app.config.suppress_callback_exceptions = True



from gama.genetic_programming.components import Fitness
from gama.genetic_programming.selection import eliminate_from_pareto
from .unit_fixtures import pset, GaussianNB, RandomForestPipeline, LinearSVC


def test_individual_length(GaussianNB, RandomForestPipeline, LinearSVC):
    assert 1 == len(list(GaussianNB.primitives))
    assert 2 == len(list(RandomForestPipeline.primitives))
    assert 1 == len(list(LinearSVC.primitives))


def test_eliminate_NSGA(GaussianNB, RandomForestPipeline, LinearSVC):
    GaussianNB.fitness = Fitness((3, -2), 0, 0, 0)
    RandomForestPipeline.fitness = Fitness((4, -2), 0, 0, 0)
    LinearSVC.fitness = Fitness((3, -1), 0, 0, 0)

    eliminated = eliminate_from_pareto(pop=[GaussianNB, RandomForestPipeline, LinearSVC], n=1)
    assert eliminated == [GaussianNB], "The element (3, -2) is dominated by both others and should be eliminated."

    # Check order independence
    eliminated = eliminate_from_pareto(pop=[RandomForestPipeline, GaussianNB, LinearSVC], n=1)
    assert eliminated == [GaussianNB], "Individual should be dominated regardless of order."



import random
from typing import List

from gama.genetic_programming.components import Primitive, Terminal, PrimitiveNode, DATA_TERMINAL


def random_terminals_for_primitive(primitive_set: dict, primitive: Primitive) -> List[Terminal]:
    """ Return a list with a Terminal matching each terminal type the Primitive requires. """
    return [random.choice(primitive_set[needed_terminal_type]) for needed_terminal_type in primitive.input]


def random_primitive_node(output_type: str, primitive_set: dict, exclude: Primitive = None) -> PrimitiveNode:
    """ Create a PrimitiveNode with a Primitive of specified output_type, with random terminals. """
    primitive = random.choice([p for p in primitive_set[output_type] if p != exclude])
    terminals = random_terminals_for_primitive(primitive_set, primitive)
    return PrimitiveNode(primitive, data_node=DATA_TERMINAL, terminals=terminals)


def create_random_expression(primitive_set: dict, min_length: int = 1, max_length: int = 3) -> PrimitiveNode:
    """ Create an PrimitiveNode with at least `min_length` and at most `max_length` chained PrimitiveNodes. """
    individual_length = random.randint(min_length, max_length)
    learner_node = random_primitive_node(output_type='prediction', primitive_set=primitive_set)
    last_primitive_node = learner_node
    for _ in range(individual_length - 1):
        primitive_node = random_primitive_node(output_type=DATA_TERMINAL, primitive_set=primitive_set)
        last_primitive_node._data_node = primitive_node
        last_primitive_node = primitive_node

    return learner_node



from typing import List, Dict

import pandas as pd
from plotly import graph_objects as go

from gama.logging.GamaReport import GamaReport


def plot_preset_graph(reports: List[GamaReport], aggregate_df: pd.DataFrame, preset: str, aggregate: str):
    if reports == []:
        return {}

    plots = []
    layout = {}
    first_metric = f'{reports[0].metrics[0]}'
    first_metric_max = f'{first_metric}_cummax'

    if preset == 'best_over_n':
        if aggregate == 'separate-line':
            plots = [individual_plot(report, 'n', first_metric_max, 'lines')
                     for report in reports]
        elif aggregate == 'aggregate':
            plots = aggregate_plot(aggregate_df, 'n', first_metric_max)
        layout = dict(
            title='Best score by iteration',
            xaxis=dict(title='n'),
            yaxis=dict(title=f'max {first_metric}'),
            hovermode='closest'
        )
    elif preset == 'best_over_time':
        if aggregate == 'separate-line':
            plots = [individual_plot(report, 'relative_end', first_metric_max, 'lines')
                     for report in reports]
        elif aggregate == 'aggregate':
            plots = aggregate_best_over_time(aggregate_df, first_metric_max)
        layout = dict(
            title=f'Best score over time',
            xaxis=dict(title='time (s)'),
            yaxis=dict(title=f'max {first_metric}'),
            hovermode='closest'
        )
    elif preset == 'size_vs_metric':
        if aggregate == 'separate-line':
            plots = [individual_plot(report, first_metric, 'length', 'markers')
                     for report in reports]
        elif aggregate == 'aggregate':
            plots = []
            for method in aggregate_df.search_method.unique():
                method_df = aggregate_df[aggregate_df.search_method == method]
                plots.append(go.Scatter(
                    x=method_df[first_metric],
                    y=method_df.length,
                    mode='markers',
                    name=method
                ))
        layout = dict(
            title=f'Size vs {first_metric}',
            xaxis=dict(title=first_metric),
            yaxis=dict(title='pipeline length'),
            hovermode='closest'
        )
    elif preset == 'number_pipeline_by_size':
        if aggregate == 'separate-line':
            for report in reports:
                size_counts = report.evaluations.length.value_counts()
                size_ratio = size_counts / len(report.individuals)
                plots.append(go.Bar(
                    x=size_ratio.index.values,
                    y=size_ratio.values,
                    name=report.name)
                )
        elif aggregate == 'aggregate':
            for method in aggregate_df.search_method.unique():
                results_for_method = aggregate_df[aggregate_df.search_method == method]
                size_counts = results_for_method.length.value_counts()
                size_ratio = size_counts / len(results_for_method)
                plots.append(go.Bar(
                    x=size_ratio.index.values,
                    y=size_ratio.values,
                    name=method)
                )
        layout = dict(
            title=f'Ratio of pipelines by size',
            xaxis=dict(title='pipeline length'),
            yaxis=dict(title='pipeline count')
        )
    elif preset == 'number_pipeline_by_learner':
        for report in reports:
            main_learners = [str(ind.main_node._primitive) for ind in report.individuals.values()]
            learner_counts = pd.Series(main_learners).value_counts()
            learner_ratio = learner_counts / len(report.individuals)
            plots.append(go.Bar(
                x=learner_ratio.index.values,
                y=learner_ratio.values,
                name=report.name)
            )
        layout = dict(
            title=f'Ratio of pipelines by learner',
            xaxis=dict(title='pipeline length'),
            yaxis=dict(title='learner')
        )
    elif preset == 'evaluation_times_dist':
        if aggregate == 'separate-line':
            for report in reports:
                time_s = report.evaluations.duration.dt.total_seconds()
                plots.append(go.Histogram(
                    x=time_s,
                    name=report.name)
                )
        elif aggregate == 'aggregate':
            for method in aggregate_df.search_method.unique():
                time_s = aggregate_df[aggregate_df.search_method == method].duration.dt.total_seconds()
                plots.append(go.Histogram(
                    x=time_s,
                    name=method)
                )
        layout = dict(
            title=f'Pipeline Evaluation Times',
            xaxis=dict(title='duration (s)'),
            yaxis=dict(title='count')
        )
    elif preset == 'n_by_rung':
        for report in reports:
            if report.search_method == 'AsynchronousSuccessiveHalving':
                count_by_rung = report.method_data.groupby(by='rung').n.count().reset_index()
                plots.append(go.Bar(
                    x=count_by_rung.rung,
                    y=count_by_rung.n,
                    name=report.name)
                )
        layout = dict(
            title=f'#Evaluations by Rung',
            xaxis=dict(title='rung'),
            yaxis=dict(title='count')
        )
    elif preset == 'time_by_rung':
        for report in reports:
            if report.search_method == 'AsynchronousSuccessiveHalving':
                duration_by_rung = report.method_data.groupby(by='rung').duration.sum().reset_index()
                duration_by_rung.duration = duration_by_rung.duration.dt.total_seconds()
                plots.append(go.Bar(
                    x=duration_by_rung.rung,
                    y=duration_by_rung.duration,
                    name=report.name)
                )
        layout = dict(
            title=f'Time spent per Rung',
            xaxis=dict(title='rung'),
            yaxis=dict(title='time (s)')
        )
    return {
        'data': plots,
        'layout': layout
    }


def individual_plot(report: GamaReport, x_axis: str, y_axis: str, mode: str):
    """

    :param report: report to pull data from
    :param x_axis: metric on the x-axis, column of report.evaluations
    :param y_axis: metric on the y-axis, column of report.evaluations
    :param mode: See `https://plot.ly/python/reference/#scatter-mode`
    :return:
        dash graph
    """
    print(len(report.evaluations[x_axis]))
    print(len(report.evaluations[y_axis]))
    return go.Scatter(
            name=f'{report.name}',
            x=report.evaluations[x_axis],
            y=report.evaluations[y_axis],
            text=[ind.short_name for ind in report.individuals.values()],
            mode=mode
        )


def aggregate_best_over_time(aggregate: pd.DataFrame, y_axis: str):
    # By method, find the best score per trace at each point in time.
    # I'm sure there is a better way to do this...
    colors = {0: 'rgba(255, 0, 0, {a})', 1: 'rgba(0, 255, 0, {a})', 2: 'rgba(0, 0, 255, {a})'}
    soft_colors = {i: color.format(a=0.2) for i, color in colors.items()}
    hard_colors = {i: color.format(a=1.0) for i, color in colors.items()}

    aggregate_data = []
    aggregate = aggregate[aggregate[y_axis] != -float('inf')]
    for color_no, method in enumerate(aggregate.search_method.unique()):
        method_agg = aggregate[aggregate.search_method == method]
        all_end_times = method_agg.relative_end.unique()
        end_time_series = pd.Series(sorted(all_end_times), name='relative_end')

        scores_over_time = pd.DataFrame(end_time_series)
        for log_no in method_agg.log_no.unique():
            log_results = method_agg[method_agg.log_no == log_no]
            best_for_log = pd.merge_asof(end_time_series, log_results[['relative_end', y_axis]], on='relative_end')
            scores_over_time[f'log{log_no}'] = best_for_log[y_axis]
        scores_over_time = scores_over_time.set_index('relative_end')
        mean_scores = scores_over_time.mean(axis=1)
        std_scores = scores_over_time.std(axis=1)
        scores_over_time['mean'] = mean_scores
        scores_over_time['std'] = std_scores

        upper_bound = go.Scatter(
            x=scores_over_time.index,
            y=scores_over_time['mean'] + scores_over_time['std'],
            mode='lines',
            marker=dict(color=soft_colors[color_no]),
            line=dict(width=0),
            fillcolor=soft_colors[color_no],
            fill='tonexty',
            showlegend=False
        )

        mean_performance = go.Scatter(
            name=method,
            x=scores_over_time.index,
            y=scores_over_time['mean'],
            mode='lines',
            line=dict(color=hard_colors[color_no]),
            fillcolor=soft_colors[color_no],
            fill='tonexty'
        )

        lower_bound = go.Scatter(
            x=scores_over_time.index,
            y=scores_over_time['mean'] - scores_over_time['std'],
            mode='lines',
            marker=dict(color=soft_colors[color_no]),
            line=dict(width=0),
            showlegend=False
        )
        aggregate_data += [lower_bound, mean_performance, upper_bound]
    return aggregate_data


def aggregate_plot(aggregate: pd.DataFrame, x_axis: str, y_axis: str):
    """ Creates an aggregate plot over multiple reports by calculating the mean and std of `y_axis` by `x_axis`.

    :param aggregate: dataframe with all evaluations
    :param x_axis: column which is grouped by before aggregating `y_axis`
    :param y_axis: column over which to calculate the mean/std.
    :return:
        Three dash Scatter objects which respectively draw the lower bound, mean and upper bound.
    """
    colors = {0: 'rgba(255, 0, 0, {a})', 1: 'rgba(0, 255, 0, {a})', 2: 'rgba(0, 0, 255, {a})'}
    soft_colors = {i: color.format(a=0.2) for i, color in colors.items()}
    hard_colors = {i: color.format(a=1.0) for i, color in colors.items()}

    # concat_df = pd.concat([report.evaluations for report in reports_to_combine])
    # concat_df = concat_df[concat_df[y_axis] != -float('inf')]
    # agg_df = concat_df.groupby(by=x_axis).agg({y_axis: ['mean', 'std']}).reset_index()
    # agg_df.columns = [x_axis, y_axis, 'std']
    aggregate_data = []
    aggregate = aggregate[aggregate[y_axis] != -float('inf')]
    for color_no, method in enumerate(aggregate.search_method.unique()):
        agg_for_method = aggregate[aggregate.search_method == method]
        agg_df = agg_for_method.groupby(by=x_axis).agg({y_axis: ['mean', 'std']}).reset_index()
        agg_df.columns = [x_axis, y_axis, 'std']

        upper_bound = go.Scatter(
            x=agg_df[x_axis],
            y=agg_df[y_axis] + agg_df['std'],
            mode='lines',
            marker=dict(color=soft_colors[color_no]),
            line=dict(width=0),
            fillcolor=soft_colors[color_no],
            fill='tonexty',
            showlegend=False
        )

        mean_performance = go.Scatter(
            name=method,
            x=agg_df[x_axis],
            y=agg_df[y_axis],
            mode='lines',
            line=dict(color=hard_colors[color_no]),
            fillcolor=soft_colors[color_no],
            fill='tonexty'
        )

        lower_bound = go.Scatter(
            x=agg_df[x_axis],
            y=agg_df[y_axis] - agg_df['std'],
            mode='lines',
            marker=dict(color=soft_colors[color_no]),
            line=dict(width=0),
            showlegend=False
        )
        aggregate_data += [lower_bound, mean_performance, upper_bound]
    return aggregate_data


"""
This submodule contains the different search methods available in GAMA.

.. note::
    You will notice an inconsistency between the signature default types and values,
    and the one in the description. The signature has `Optional` values which default
    to `None`. If left at `None`, they will default to the value specified in the description,
    which may be a value based on data characteristics.

"""

from gama.search_methods.asha import AsynchronousSuccessiveHalving
from gama.search_methods.async_ea import AsyncEA
from gama.search_methods.random_search import RandomSearch
from gama.search_methods.base_search import _check_base_search_hyperparameters


__all__ = ['AsynchronousSuccessiveHalving', 'AsyncEA', 'RandomSearch']



from datetime import datetime
import logging
import os
import pickle
import time
from typing import Iterable
import uuid

import stopit
from sklearn.preprocessing import OneHotEncoder
from sklearn.model_selection import cross_val_predict, ShuffleSplit
from sklearn.pipeline import Pipeline

from gama.utilities.metrics import Metric
from gama.genetic_programming.components import Individual, PrimitiveNode, Fitness
from gama.logging.utility_functions import MultiprocessingLogger
from gama.logging.machine_logging import TOKENS, log_event

log = logging.getLogger(__name__)


def primitive_node_to_sklearn(primitive_node: PrimitiveNode) -> object:
    hyperparameters = {terminal.output: terminal.value for terminal in primitive_node._terminals}
    return primitive_node._primitive.identifier(**hyperparameters)


def compile_individual(individual: Individual, parameter_checks=None, preprocessing_steps: Iterable[object]=None) -> Pipeline:
    steps = [(str(i), primitive_node_to_sklearn(primitive)) for i, primitive in enumerate(individual.primitives)]
    if preprocessing_steps:
        steps = steps + [(str(i), step) for (i, step) in enumerate(reversed(preprocessing_steps), start=len(steps))]
    return Pipeline(list(reversed(steps)))


def cross_val_predict_score(estimator, X, y_train, metrics=None, **kwargs):
    """ Return both the predictions and score of the estimator trained on the data given the cv strategy.

    :param y_train: target in appropriate format for training (typically (N,))
    """
    if not all(isinstance(metric, Metric) for metric in metrics):
        raise ValueError('All `metrics` must be an instance of `metrics.Metric`, is {}.'
                         .format([type(metric) for metric in metrics]))

    predictions_are_probabilities = any(metric.requires_probabilities for metric in metrics)
    method = 'predict_proba' if predictions_are_probabilities else 'predict'
    predictions = cross_val_predict(estimator, X, y_train, method=method, **kwargs)

    if predictions.ndim == 2 and predictions.shape[1] == 1:
        predictions = predictions.squeeze()

    scores = []
    for metric in metrics:
        if metric.requires_probabilities:
            # `predictions` are of shape (N,K) and the ground truth should be formatted accordingly
            y_ohe = OneHotEncoder().fit_transform(y_train.values.reshape(-1, 1)).toarray()
            scores.append(metric.maximizable_score(y_ohe, predictions))
        elif predictions_are_probabilities:
            # Metric requires no probabilities, but probabilities were predicted.
            scores.append(metric.maximizable_score(y_train, predictions.argmax(axis=1)))
        else:
            # No metric requires probabilities, so `predictions` is an array of labels.
            scores.append(metric.maximizable_score(y_train, predictions))

    return predictions, scores


def object_is_valid_pipeline(o):
    """ Determines if object behaves like a valid scikit-learn pipeline (it must have `fit`, `predict` and `steps`). """
    return (o is not None and
            hasattr(o, 'fit') and
            hasattr(o, 'predict') and
            hasattr(o, 'steps'))


def evaluate_individual(individual: Individual, evaluate_pipeline_length, *args, **kwargs):
    (scores, start_datetime, wallclock_time, process_time) = evaluate_pipeline(individual.pipeline, *args, **kwargs)
    if evaluate_pipeline_length:
        scores = (*scores, -len(individual.primitives))
    individual.fitness = Fitness(scores, start_datetime, wallclock_time, process_time)
    return individual


def evaluate_pipeline(pl, X, y_train, timeout, deadline, metrics='accuracy', cv=5, cache_dir=None, logger=None, subsample=None):
    """ Evaluates a pipeline used k-Fold CV. """
    if not logger:
        logger = log

    if not object_is_valid_pipeline(pl):
        return ValueError('Pipeline is not valid. Must not be None and have `fit`, `predict` and `steps`.')

    draw_subsample = (isinstance(subsample, int) and subsample < len(y_train))
    scores = tuple([float('-inf')] * len(metrics))
    start_datetime = datetime.now()
    start = time.process_time()

    time_to_deadline = deadline - time.time()
    timeout = min(timeout, time_to_deadline)
    with stopit.ThreadingTimeout(timeout) as c_mgr:
        try:
            if draw_subsample:
                idx, _ = next(ShuffleSplit(n_splits=1, train_size=subsample, random_state=0).split(X))
                X, y_train = X.iloc[idx, :], y_train[idx]

            prediction, scores = cross_val_predict_score(pl, X, y_train, cv=cv, metrics=metrics)
        except stopit.TimeoutException:
            # score not actually unused, because exception gets caught by the context manager.
            raise
        except KeyboardInterrupt:
            raise
        except Exception as e:
            if isinstance(logger, MultiprocessingLogger):
                logger.debug('{} encountered while evaluating pipeline.'.format(type(e)))
            else:
                logger.debug('{} encountered while evaluating pipeline.'.format(type(e)), exc_info=True)

            single_line_pipeline = str(pl).replace('\n', '')
            log_event(logger, TOKENS.EVALUATION_ERROR, start_datetime, single_line_pipeline, type(e), e)

    if cache_dir and -float("inf") not in scores and not draw_subsample:
        pl_filename = str(uuid.uuid4())
        try:
            with open(os.path.join(cache_dir, pl_filename + '.pkl'), 'wb') as fh:
                pickle.dump((pl, prediction, scores), fh)
        except FileNotFoundError:
            log.debug("File not found while saving predictions. This can happen in the multi-process case if the "
                      "cache gets deleted within `max_eval_time` of the end of the search process.", exc_info=True)

    process_time = time.process_time() - start
    wallclock_time = (datetime.now() - start_datetime).total_seconds()

    if c_mgr.state == c_mgr.INTERRUPTED:
        # A TimeoutException was raised, but not by the context manager.
        # This indicates that the outer context manager (the ea) timed out.
        logger.info("Outer-timeout during evaluation of {}".format(pl))
        raise stopit.utils.TimeoutException()

    if not c_mgr:
        # For now we treat an eval timeout the same way as e.g. NaN exceptions and use the default score.
        logger.info('Timeout encountered while evaluating pipeline.')
        single_line_pipeline = ''.join(str(pl).split('\n'))
        log_event(logger, TOKENS.EVALUATION_TIMEOUT, start_datetime, single_line_pipeline)
        logger.debug("Timeout after {}s: {}".format(timeout, pl))

    fitness_values = (scores, start_datetime, wallclock_time, process_time)
    return fitness_values



import time


class Stopwatch:
    """ A context manager that keeps track of wall clock time spent. """

    def __init__(self, timing_function=time.time):
        """

        Parameters
        ----------
        timing_function: Callable (default=time.time)
            The function used to measure time, e.g. time.time or time.process_time
        """
        self._is_running = False
        self._get_time = timing_function
        self._start = 0
        self._end = 0

    def __enter__(self):
        self._is_running = True
        self._start = self._get_time()
        return self

    def __exit__(self, *args):
        self._end = self._get_time()
        self._is_running = False
        return False  # do not suppress any exceptions.

    @property
    def elapsed_time(self):
        """ If used in with-statement, start since with-statement. If used after, time spent during with statement."""
        if self._is_running:
            return self._get_time() - self._start
        else:
            return self._end - self._start



from sklearn.svm import LinearSVC
from sklearn.datasets import load_iris
from gama.postprocessing.ensemble import fit_and_weight


def test_fit_and_weight():
    x, y = load_iris(return_X_y=True)
    good_estimator = LinearSVC()
    bad_estimator = LinearSVC(penalty='l1')  # Not supported with default squared hinge loss solving the dual problem

    _, w = fit_and_weight((good_estimator, x, y, 1))
    assert 1 == w
    _, w = fit_and_weight((bad_estimator, x, y, 1))
    assert 0 == w



""" This module defines the built-in metrics for GAMA.

The valid metrics depend on the type of task. Many scikit-learn metrics are available.
For classification, the following metrics are available:

    - accuracy, roc_auc, average_precision, log_loss
    - micro, macro, weighted and sample averages for precision, recall and f1 (e.g. f1_macro, precision_micro)

For regression, the following metrics are available:

    explained_variance, r2, median_absolute_error, mean_squared_error, mean_squared_log_error, mean_absolute_error

Whether to minimize or maximize is determined automatically.
"""
from enum import Enum
from functools import partial
from typing import Callable, Iterable, Tuple, Union, Dict

import numpy as np
import pandas as pd
from sklearn import metrics

"""
Scikit-learn does not have an option to return predictions and score at the same time. Furthermore, the only string
interpretation of scoring functions automatically make 'scorers' which train the model internally, also throwing
away any predictions. So we need to make our own conversion of scoring string to function, predict, score, and return
both. Construction of metric_strings copied with minor modifications from SCORERS of scikit-learn. See also:

1. https://github.com/scikit-learn/scikit-learn/blob/a24c8b464d094d2c468a16ea9f8bf8d42d949f84/sklearn/metrics/scorer.py#L530
2. https://stackoverflow.com/questions/41003897/scikit-learn-cross-validates-score-and-predictions-at-one-go
"""

# name: (Score function, bool: requires_probabilities, bool: maximizing optimizes)
classification_metrics = dict(
    accuracy=(metrics.accuracy_score, False, True),
    roc_auc=(metrics.roc_auc_score, True, True),
    average_precision=(metrics.average_precision_score, True, True),
    log_loss=(metrics.log_loss, True, False),
    neg_log_loss=(metrics.log_loss, True, False)
)

# Below is also based on scikit-learn code:
for name, score_fn in [('precision', metrics.precision_score),
                       ('recall', metrics.recall_score),
                       ('f1', metrics.f1_score)]:
    classification_metrics[name] = (score_fn, False, True)
    for average in ['macro', 'micro', 'samples', 'weighted']:
        qualified_name = '{0}_{1}'.format(name, average)
        qualified_score_fn = partial(score_fn, average=average)
        classification_metrics[qualified_name] = (qualified_score_fn, False, True)

regression_metrics: Dict[str, Tuple[Callable, bool, bool]] = dict(
    explained_variance=(metrics.explained_variance_score, False, True),
    r2=(metrics.r2_score, False, True),
    neg_mean_absolute_error=(metrics.mean_absolute_error, False, False),
    mean_absolute_error=(metrics.mean_absolute_error, False, False),
    neg_mean_squared_log_error=(metrics.mean_squared_log_error, False, False),
    mean_squared_log_error=(metrics.mean_squared_log_error, False, False),
    neg_median_absolute_error=(metrics.median_absolute_error, False, False),
    median_absolute_error=(metrics.median_absolute_error, False, False),
    neg_mean_squared_error=(metrics.mean_squared_error, False, False),
    mean_squared_error=(metrics.mean_squared_error, False, False)
)

all_metrics = {**classification_metrics, **regression_metrics}


class MetricType(Enum):
    """ Metric types supported by GAMA. """
    CLASSIFICATION: int = 1  #: discrete target
    REGRESSION: int = 2  #: continuous target


class Metric:
    """ A wrapper for a scoring function to provide additional meta-data. """

    def __init__(self, metric_name: str,
                 score_function: Callable,
                 requires_probabilities: bool,
                 maximize: bool,
                 task_type: MetricType):
        """
        Parameters
        ----------
        metric_name: str
            Name of the metric.
        score_function: Callable
            Takes either predicted targets or predicted probabilities and the true target values to calculate a score.
        requires_probabilities: bool
            For classification metrics only, describes if the metric should be computed with predicted probabilities.
        maximize: bool
            If True, indicates that maximizing the score corresponds to better models.
        task_type: MetricType
            Type of the machine learning task.
        """
        self.name = metric_name
        self._score_function = score_function
        self.requires_probabilities = requires_probabilities
        self._optimize_modifier = 1 if maximize else -1
        self.task_type = task_type

    def score(self, y_true, predictions) -> float:
        """ Score the predictions based on the metric.

        :param y_true: numpy array of shape (N,K) if metric relies on class probabilities, (N,) otherwise.
        :param predictions: numpy array of shape (N,K) if metric relies on class probabilities, (N,) otherwise.
        :return: score of predictions according to the metric.
        """
        # Scikit-learn metrics can be very flexible with their input, interpreting a list as class labels for one
        # metric, while interpreting it as class probability for the positive class for another.
        # We want to force clear and early errors to avoid accidentally working with the wrong data/interpretation.
        if not isinstance(y_true, (np.ndarray, pd.Series, pd.DataFrame)):
            raise TypeError('y_true must be a numpy array, pandas series or pandas dataframe.')
        if not isinstance(predictions, (np.ndarray, pd.Series, pd.DataFrame)):
            raise TypeError('predictions must be a numpy array, pandas series or pandas dataframe.')
        return self._score_function(y_true, predictions)

    def maximizable_score(self, y_true, predictions):
        """ Calculates the score, but negated if necessary so that maximizing is always better. """
        return self._optimize_modifier * self.score(y_true, predictions)

    @classmethod
    def from_string(cls, metric_name: str) -> 'Metric':
        if metric_name in regression_metrics:
            task_type = MetricType.REGRESSION
        elif metric_name in classification_metrics:
            task_type = MetricType.CLASSIFICATION
        else:
            raise ValueError('Metric not known: {}.'.format(metric_name))

        score_function, requires_probabilities, should_maximize = all_metrics[metric_name]
        return cls(metric_name, score_function, requires_probabilities, should_maximize, task_type)


def scoring_to_metric(scoring: Union[str, Metric, Iterable[str], Iterable[Metric]]) -> Tuple[Metric]:
    if isinstance(scoring, str):
        return tuple([Metric.from_string(scoring)])
    elif isinstance(scoring, Metric):
        return tuple([scoring])
    elif (isinstance(scoring, Iterable)
            and all([isinstance(scorer, Metric) or isinstance(scorer, str) for scorer in scoring])):
        return tuple([scorer if isinstance(scorer, Metric) else Metric.from_string(scorer) for scorer in scoring])
    else:
        raise ValueError("scoring must be a string, Metric or Iterable (of strings or Metrics).")



import numpy as np
import pandas as pd
from gama.data import arff_to_pandas, X_y_from_arff

NUMERIC_TYPES = [np.int, np.int32, np.int64, np.float]


def test_arff_to_pandas():
    # https://www.openml.org/d/23380
    dataframe = arff_to_pandas('tests/data/openml_d_23380.arff')
    assert isinstance(dataframe, pd.DataFrame)
    assert (2796, 35) == dataframe.shape
    assert 68100 == dataframe.isnull().sum().sum()
    assert 32 == sum([dtype in NUMERIC_TYPES for dtype in dataframe.dtypes])
    assert 3 == sum([dtype.name == 'category' for dtype in dataframe.dtypes])


def test_X_y_from_arff():
    # https://www.openml.org/d/23380
    x, y = X_y_from_arff('tests/data/openml_d_23380.arff', split_column='TR')

    assert isinstance(x, pd.DataFrame)
    assert (2796, 34) == x.shape
    assert 68100 == x.isnull().sum().sum()
    assert 32 == sum([dtype in NUMERIC_TYPES for dtype in x.dtypes])
    assert 2 == sum([dtype.name == 'category' for dtype in x.dtypes])

    assert isinstance(y, pd.Series)
    assert (2796,) == y.shape
    assert 0 == y.isnull().sum()
    assert 6 == len(y.dtype.categories)



import time
import pytest
from gama.utilities.generic.timekeeper import TimeKeeper


ROUND_ERROR = 0.02


def test_timekeeper_total_time_remaning_error_if_total_time_zero():
    """ Ensure `total_time_remaining` is unavailable if `total_time` is not set. """
    timekeeper = TimeKeeper()
    with pytest.raises(RuntimeError):
        _ = timekeeper.total_time_remaining


def test_timekeeper_stopwatch_normal_behavior():
    """ Ensure normal stopwatch functionality for stopwatch returned by context manager. """
    timekeeper = TimeKeeper()
    with timekeeper.start_activity('test activity', time_limit=3) as sw:
        assert pytest.approx(0, abs=ROUND_ERROR) == sw.elapsed_time
        assert pytest.approx(0, abs=ROUND_ERROR) == timekeeper.current_activity_time_elapsed
        assert sw._is_running
        assert pytest.approx(3, abs=ROUND_ERROR) == timekeeper.current_activity_time_left
        time.sleep(1)
        assert pytest.approx(1, abs=ROUND_ERROR) == sw.elapsed_time
        assert pytest.approx(1, abs=ROUND_ERROR) == timekeeper.current_activity_time_elapsed
        assert pytest.approx(2, abs=ROUND_ERROR) == timekeeper.current_activity_time_left
        assert sw._is_running

    time.sleep(1)
    assert not sw._is_running
    assert pytest.approx(1, abs=ROUND_ERROR) == sw.elapsed_time

    with pytest.raises(RuntimeError) as error:
        timekeeper.current_activity_time_elapsed
    assert "No activity in progress." in str(error.value)

    with pytest.raises(RuntimeError) as error:
        timekeeper.current_activity_time_left
    assert "No activity in progress." in str(error.value)


def test_timekeeper_total_remaining_time():
    """ Ensure total remaining time is correct across activities. """
    total_time = 10
    timekeeper = TimeKeeper(total_time=total_time)
    assert timekeeper.total_time_remaining == total_time

    activity_length = 1
    with timekeeper.start_activity('part one'):
        time.sleep(activity_length)

    time.sleep(1)
    assert pytest.approx(activity_length, abs=ROUND_ERROR) == timekeeper.activities[-1].stopwatch.elapsed_time
    assert pytest.approx(total_time - activity_length, abs=ROUND_ERROR) == timekeeper.total_time_remaining



import pytest

import gama


def test_reproducible_initialization():
    g1 = gama.GamaClassifier(random_state=1, keep_analysis_log=None)
    pop1 = [g1._operator_set.individual() for _ in range(10)]

    g2 = gama.GamaClassifier(random_state=1, keep_analysis_log=None)
    pop2 = [g2._operator_set.individual() for _ in range(10)]
    for ind1, ind2 in zip(pop1, pop2):
        assert ind1.pipeline_str() == ind2.pipeline_str(), "The initial population should be reproducible."


def test_gama_fail_on_invalid_hyperparameter_values():
    # `delete_cache` is only called when the initialization failed to raise a ValueError.
    # In that case, we would need to delete the created cache directory. Otherwise, it would not even be created.
    with pytest.raises(ValueError) as e:
        gama.GamaClassifier(max_total_time=0).delete_cache()
    assert "max_total_time should be integer greater than zero" in str(e.value)

    with pytest.raises(ValueError) as e:
        gama.GamaClassifier(max_total_time=None).delete_cache()
    assert "max_total_time should be integer greater than zero" in str(e.value)

    with pytest.raises(ValueError) as e:
        gama.GamaClassifier(max_eval_time=0).delete_cache()
    assert "max_eval_time should be None or integer greater than zero" in str(e.value)

    with pytest.raises(ValueError) as e:
        gama.GamaClassifier(n_jobs=-2).delete_cache()
    assert "n_jobs should be -1 or positive integer" in str(e.value)



from typing import List
from .terminal import DATA_TERMINAL, Terminal
from .primitive import Primitive


class PrimitiveNode:
    """ An instantiation for a Primitive with specific Terminals.

    Parameters
    ----------
    primitive: Primitive
        The Primitive type of this PrimitiveNode.
    data_node: PrimitiveNode
        The PrimitiveNode that specifies all preprocessing before this PrimitiveNode.
    terminals: List[Terminal]
        A list of terminals matching the `primitive`.
    """

    def __init__(self, primitive: Primitive, data_node: 'PrimitiveNode', terminals: List[Terminal]):
        self._primitive = primitive
        self._data_node = data_node
        self._terminals = sorted(terminals, key=lambda t: str(t))

    def __str__(self):
        """ Recursively stringify all primitive nodes (primitive and hyperparameters).

        Examples: - "GaussianNB(data)"
                  - "BernoulliNB(data, alpha=1.0)"
                  - "BernoulliNB(FastICA(data, tol=0.5), alpha=1.0)"
        """
        if self._terminals:
            terminal_str = ", ".join([repr(terminal) for terminal in self._terminals])
            return "{}({}, {})".format(self._primitive, str(self._data_node), terminal_str)
        else:
            return "{}({})".format(self._primitive, str(self._data_node))

    def copy(self):
        """ Make a shallow copy w.r.t. Primitive/Terminal (they are immutable), but deep w.r.t. PrimitiveNodes. """
        data_node_copy = self._data_node if self._data_node == DATA_TERMINAL else self._data_node.copy()
        return PrimitiveNode(primitive=self._primitive, data_node=data_node_copy, terminals=self._terminals.copy())

    @classmethod
    def from_string(cls, string: str, primitive_set: dict):
        """ Create a PrimitiveNode from string formatted like PrimitiveNode.__str__

        Parameters
        ----------
        string: str
            A string formatted similar to PrimitiveNode.__str__
        primitive_set: dict
            The dictionary defining all Terminals and Primitives.

        Returns
        -------
        PrimitiveNode
            The PrimitiveNode as defined the string.
        """
        # General form is A(B(C(data[, C.param=value, ...])[, B.param=value, ...])[, A.param=value, ...])
        # below assumes that left parenthesis is never part of a parameter name or value.
        primitives = string.split('(')[:-1]
        terminal_start_index = string.index(DATA_TERMINAL)
        terminals_string = string[terminal_start_index + len(DATA_TERMINAL):]
        terminal_sets = terminals_string.split(')')[:-1]

        last_node = DATA_TERMINAL
        for primitive_string, terminal_set in zip(reversed(primitives), terminal_sets):
            primitive = find_primitive(primitive_set, primitive_string)
            if terminal_set == '':
                terminals = []
            else:
                terminal_set = terminal_set[2:]  # 2 is because string starts with ', '
                terminals = [find_terminal(primitive_set, terminal_string)
                             for terminal_string in terminal_set.split(', ')]
            if not all([required_terminal in map(lambda t: t.identifier, terminals)
                        for required_terminal in primitive.input]):
                missing = [required_terminal for required_terminal in primitive.input
                           if required_terminal not in map(lambda t: t.identifier, terminals)]
                raise ValueError("Individual does not define all required terminals for primitive {}. Missing: {}."
                                 .format(primitive, missing))
            last_node = cls(primitive, last_node, terminals)

        return last_node


def find_primitive(primitive_set: dict, primitive_string: str) -> Primitive:
    """ Find the Primitive that matches `primitive_string` in `primitive_set`. """
    all_primitives = primitive_set[DATA_TERMINAL] + primitive_set['prediction']
    try:
        return [p for p in all_primitives if repr(p) == primitive_string][0]
    except IndexError:
        raise IndexError(f"Could not find Primitive of type '{primitive_string}'.")



def find_terminal(primitive_set: dict, terminal_string: str) -> Terminal:
    """ Find the Terminal that matches `terminal_string` in `primitive_set`. """
    terminal_return_type, terminal_value = terminal_string.split('=')
    try:
        return [t for t in primitive_set[terminal_return_type] if repr(t) == terminal_string][0]
    except IndexError:
        raise IndexError(f"Could not find Terminal of type '{terminal_string}'.")



from collections import defaultdict
import pytest

import numpy as np

from gama.genetic_programming.components import Individual
from gama.genetic_programming.mutation import mut_replace_terminal, mut_replace_primitive, random_valid_mutation_in_place
from gama.genetic_programming.compilers.scikitlearn import compile_individual
from .unit_fixtures import pset, GaussianNB, RandomForestPipeline, LinearSVC

def test_mut_replace_terminal(RandomForestPipeline, pset):
    """ Tests if mut_replace_terminal replaces exactly one terminal. """
    _test_mutation(RandomForestPipeline, mut_replace_terminal, _mut_replace_terminal_is_applied, pset)


def test_mut_replace_terminal_none_available(GaussianNB, pset):
    """ Tests if mut_replace_terminal raises an exception if no valid mutation is possible. """
    with pytest.raises(ValueError) as error:
        mut_replace_terminal(GaussianNB, pset)

    assert "Individual has no terminals or no terminals suitable for mutation." in str(error.value)


def test_mut_replace_primitive_len_1(LinearSVC, pset):
    """ Tests if mut_replace_primitive replaces exactly one primitive. """
    _test_mutation(LinearSVC, mut_replace_primitive, _mut_replace_primitive_is_applied, pset)


def test_mut_replace_primitive_len_2(RandomForestPipeline, pset):
    """ Tests if mut_replace_primitive replaces exactly one primitive. """
    _test_mutation(RandomForestPipeline, mut_replace_primitive, _mut_replace_primitive_is_applied, pset)


def test_random_valid_mutation_with_all(RandomForestPipeline, pset):
    """ Test if a valid mutation is applied at random.

    I am honestly not sure of the best way to test this.
    Because of the random nature, we repeat this enough times to ensure each mutation is tried with
    probability >0.99999.
    """

    applied_mutation = defaultdict(int)
    N = _min_trials(n_mutations=4)

    for i in range(N):
        ind_clone = RandomForestPipeline.copy_as_new()
        random_valid_mutation_in_place(ind_clone, pset)
        if _mutShrink_is_applied(RandomForestPipeline, ind_clone)[0]:
            applied_mutation['shrink'] += 1
        elif _mutInsert_is_applied(RandomForestPipeline, ind_clone)[0]:
            applied_mutation['insert'] += 1
        elif _mut_replace_terminal_is_applied(RandomForestPipeline, ind_clone)[0]:
            applied_mutation['terminal'] += 1
        elif _mut_replace_primitive_is_applied(RandomForestPipeline, ind_clone)[0]:
            applied_mutation['primitive'] += 1
        else:
            assert False, "No mutation (or one that is unaccounted for) is applied."

    assert all([n > 0 for (mut, n) in applied_mutation.items()])


def test_random_valid_mutation_without_shrink(LinearSVC, pset):
    """ Test if a valid mutation is applied at random.

    I am honestly not sure of the best way to test this.
    Because of the random nature, we repeat this enough times to ensure each mutation is tried with
    probability >0.99999.
    """

    applied_mutation = defaultdict(int)
    N = _min_trials(n_mutations=3)

    for i in range(N):
        ind_clone = LinearSVC.copy_as_new()
        random_valid_mutation_in_place(ind_clone, pset)
        if _mutInsert_is_applied(LinearSVC, ind_clone)[0]:
            applied_mutation['insert'] += 1
        elif _mut_replace_terminal_is_applied(LinearSVC, ind_clone)[0]:
            applied_mutation['terminal'] += 1
        elif _mut_replace_primitive_is_applied(LinearSVC, ind_clone)[0]:
            applied_mutation['primitive'] += 1
        else:
            assert False, "No mutation (or one that is unaccounted for) is applied."

    assert all([n > 0 for (mut, n) in applied_mutation.items()])


def test_random_valid_mutation_without_terminal(GaussianNB, pset):
    """ Test if a valid mutation is applied at random.

    I am honestly not sure of the best way to test this.
    Because of the random nature, we repeat this enough times to ensure each mutation is tried with
    probability >0.99999.
    """
    # The tested individual contains no terminals and one primitive,
    # and thus is not eligible for replace_terminal and mutShrink.
    applied_mutation = defaultdict(int)
    N = _min_trials(n_mutations=2)

    for i in range(N):
        ind_clone = GaussianNB.copy_as_new()
        random_valid_mutation_in_place(ind_clone, pset)
        if _mutInsert_is_applied(GaussianNB, ind_clone)[0]:
            applied_mutation['insert'] += 1
        elif _mut_replace_primitive_is_applied(GaussianNB, ind_clone)[0]:
            applied_mutation['primitive'] += 1
        else:
            assert False, "No mutation (or one that is unaccounted for) is applied."

    assert all([n > 0 for (mut, n) in applied_mutation.items()])


def _min_trials(n_mutations, max_error_rate=0.0001):
    return int(np.ceil(np.log(max_error_rate) / np.log((n_mutations - 1) / n_mutations)))


def _mutShrink_is_applied(original, mutated):
    """ Checks if mutation was caused by `mut_shrink`.

    :param original: the pre-mutation individual
    :param mutated:  the post-mutation individual
    :return: (bool, str). If mutation was caused by function, True. False otherwise.
        str is a message explaining why mutation is not caused by function.
    """
    if len(list(original.primitives)) <= len(list(mutated.primitives)):
        return (False, "Number of primitives should be strictly less, was {} is {}."
                .format(len(list(original.primitives)), len(list(mutated.primitives))))
    return (True, None)


def _mutInsert_is_applied(original, mutated):
    """ Checks if mutation was caused by `mut_insert`.

    :param original: the pre-mutation individual
    :param mutated:  the post-mutation individual
    :return: (bool, str). If mutation was caused by function, True. False otherwise.
        str is a message explaining why mutation is not caused by function.
    """
    if len(list(original.primitives)) >= len(list(mutated.primitives)):
        return (False, "Number of primitives should be strictly greater, was {} is {}."
                .format(len(list(original.primitives)), len(list(mutated.primitives))))
    return (True, None)


def _mut_replace_terminal_is_applied(original, mutated):
    """ Checks if mutation was caused by `gama.ea.mutation.mut_replace_terminal`.

    :param original: the pre-mutation individual
    :param mutated:  the post-mutation individual
    :return: (bool, str). If mutation was caused by function, True. False otherwise.
        str is a message explaining why mutation is not caused by function.
    """
    if len(list(original.terminals)) != len(list(mutated.terminals)):
        return (False, "Number of terminals should be unchanged, was {} is {}."
                .format(len(list(original.terminals)), len(list(mutated.terminals))))

    replaced_terminals = [t1 for t1, t2 in zip(original.terminals, mutated.terminals) if str(t1) != str(t2)]
    if len(replaced_terminals) != 1:
        return (False, "Expected 1 replaced Terminal, found {}.".format(len(replaced_terminals)))
    return (True, None)


def _mut_replace_primitive_is_applied(original, mutated):
    """ Checks if mutation was caused by `gama.ea.mutation.mut_replace_primitive`.

    :param original: the pre-mutation individual
    :param mutated:  the post-mutation individual
    :return: (bool, str). If mutation was caused by function, True. False otherwise.
        str is a message explaining why mutation is not caused by function.
    """
    if len(list(original.primitives)) != len(list(mutated.primitives)):
        return (False, "Number of primitives should be unchanged, was {} is {}."
                .format(len(list(original.primitives)), len(list(mutated.primitives))))

    replaced_primitives = [p1 for p1, p2 in zip(original.primitives, mutated.primitives)
                           if str(p1._primitive) != str(p2._primitive)]
    if len(replaced_primitives) != 1:
        return (False, "Expected 1 replaced Primitive, found {}.".format(len(replaced_primitives)))
    return (True, None)


def _test_mutation(individual: Individual, mutation, mutation_check, pset):
    """ Test if an individual mutated by `mutation` passes `mutation_check` and compiles.

    :param individual: The individual to be mutated.
    :param mutation: function: ind -> (ind,). Should mutate the individual
    :param mutation_check: function: (ind1, ind2)->(bool, str).
       A function to check if ind2 could have been created by `mutation(ind1)`, see above functions.
    """
    ind_clone = individual.copy_as_new()
    mutation(ind_clone, pset)

    applied, message = mutation_check(individual, ind_clone)
    assert applied, message

    # Should be able to compile the individual, will raise an Exception if not.
    compile_individual(ind_clone, pset)




from sklearn.naive_bayes import BernoulliNB, GaussianNB

from gama.configuration.parser import merge_configurations


def test_merge_configuration():
    """ Test merging two simple configurations works as expected. """

    one = {
        'alpha': [0, 1],
        BernoulliNB: {
            'fit_prior': [True, False]
        }
    }

    two = {
        'alpha': [0, 2],
        GaussianNB: {
            'fit_prior': [True, False]
        }
    }

    expected_merged = {
        'alpha': [0, 1, 2],
        GaussianNB: {
            'fit_prior': [True, False]
        },
        BernoulliNB: {
            'fit_prior': [True, False]
        }
    }

    actual_merged = merge_configurations(one, two)
    assert expected_merged == actual_merged



from functools import partial
import logging
import math
from typing import List, Optional, Dict, Tuple, Any, Callable

import pandas as pd
import stopit

from gama.genetic_programming.operator_set import OperatorSet
from gama.search_methods.base_search import BaseSearch
from gama.logging.machine_logging import TOKENS, log_event
from gama.utilities.generic.async_evaluator import AsyncEvaluator
from gama.genetic_programming.components.individual import Individual

"""
TODO:
 - instead of list, use a min-heap by rung.
 - promoted pipelines as set and set-intersection to determine promotability?
"""

log = logging.getLogger(__name__)
ASHA_LOG_TOKEN = 'ASHA'


class AsynchronousSuccessiveHalving(BaseSearch):
    """ Asynchronous Halving Algorithm by Li et al. (paper: https://arxiv.org/abs/1810.05934)

    Parameters
    ----------
    reduction_factor: int, optional (default=3)
        Reduction factor of candidates between each rung.
    minimum_resource: int, optional (default=100)
        Number of samples to use in the lowest rung.
    maximum_resource: int, optional (default=number of samples in the dataset)
        Number of samples to use in the top rung. This should not exceed the number of samples in the data.
    minimum_early_stopping_rate: int (default=1)
        Number of lowest rungs to skip.
    """

    def __init__(self,
                 reduction_factor: Optional[int] = None,
                 minimum_resource: Optional[int] = None,
                 maximum_resource: Optional[int] = None,
                 minimum_early_stopping_rate: Optional[int] = None):
        super().__init__()
        # maps hyperparameter -> (set value, default)
        self._hyperparameters: Dict[str, Tuple[Any, Any]] = dict(
            reduction_factor=(reduction_factor, 3),
            minimum_resource=(minimum_resource, 100),
            maximum_resource=(maximum_resource, 100_000),
            minimum_early_stopping_rate=(minimum_early_stopping_rate, 1),
        )
        self.output = []

    def dynamic_defaults(self, x: pd.DataFrame, y: pd.DataFrame, time_limit: int):
        # `maximum_resource` is the number of samples used in the highest rung.
        # this typically should be the number of samples in the (training) dataset.
        self._overwrite_hyperparameter_default('maximum_resource', len(y))

    def search(self, operations: OperatorSet, start_candidates: List[Individual]):
        self.output = asha(operations, start_candidates=start_candidates, **self.hyperparameters)


def asha(operations: OperatorSet,
         start_candidates: List[Individual],
         reduction_factor: int = 3,
         minimum_resource: int = 100,
         maximum_resource: int = 100_000,
         minimum_early_stopping_rate: int = 1,
         maximum_max_rung_evaluations: Optional[int] = None) -> List[Individual]:
    """ Asynchronous Halving Algorithm by Li et al. (paper: https://arxiv.org/abs/1810.05934)

    Parameters
    ----------
    operations: OperatorSet
        An operator set with `evaluate` and `individual` functions.
    start_candidates: List[Individual]
        A list which contains the set of best found individuals during search.
    reduction_factor: int (default=3)
        Reduction factor of candidates between each rung.
    minimum_resource: int (default=100)
        Number of samples to use in the lowest rung.
    maximum_resource: int (default=100_000)
        Number of samples to use in the top rung. This should not exceed the number of samples in the data.
    minimum_early_stopping_rate: int (default=1)
        Number of lowest rungs to skip.
    maximum_max_rung_evaluations: Optional[int] (default=None)
        Maximum number of individuals to evaluate on the max rung.
        If None, the algorithm will be run indefinitely.

    Returns
    -------
    List[Individual]
        All individuals of the highest rung in which at least one individual has been evaluated.
    """
    evaluate = partial(evaluate_on_rung, evaluate_individual=operations.evaluate)

    # Note that here we index the rungs by all possible rungs (0..ceil(log_eta(R/r))), and ignore the first
    # minimum_early_stopping_rate rungs. This contrasts the paper where rung 0 refers to the first used one.
    max_rung = math.ceil(math.log(maximum_resource/minimum_resource, reduction_factor))
    rungs = range(minimum_early_stopping_rate, max_rung + 1)
    resource_for_rung = {rung: min(minimum_resource * (reduction_factor ** rung), maximum_resource) for rung in rungs}

    # Should we just use lists of lists/heaps instead?
    individuals_by_rung = {rung: [] for rung in reversed(rungs)}  # Highest rungs first is how we typically iterate them
    promoted_individuals = {rung: [] for rung in reversed(rungs)}

    def get_job():
        for rung, individuals in list(individuals_by_rung.items())[1:]:
            # This is not in the paper code but is derived from fig 2b
            n_to_promote = math.floor(len(individuals) / reduction_factor)
            if n_to_promote - len(promoted_individuals[rung]) > 0:
                # Problem: equal loss falls back on comparison of individual
                candidates = list(sorted(individuals, key=lambda t: t[0], reverse=True))[:n_to_promote]
                promotable = [candidate for candidate in candidates if candidate not in promoted_individuals[rung]]
                if len(promotable) > 0:
                    promoted_individuals[rung].append(promotable[0])
                    return promotable[0][1], rung + 1

        if start_candidates is not None and len(start_candidates) > 0:
            return start_candidates.pop(), minimum_early_stopping_rate
        else:
            return operations.individual(), minimum_early_stopping_rate

    try:
        with AsyncEvaluator() as async_:
            def start_new_job():
                individual, rung = get_job()
                time_penalty_for_rung = resource_for_rung[rung] / max(resource_for_rung.values())
                async_.submit(evaluate, individual, rung,
                              subsample=resource_for_rung[rung],
                              timeout=(10 + (time_penalty_for_rung * 600)))

            for _ in range(8):
                start_new_job()

            while ((maximum_max_rung_evaluations is None)
                   or (len(individuals_by_rung[max_rung]) < maximum_max_rung_evaluations)):
                future = operations.wait_next(async_)
                if future.result is not None:
                    individual, loss, rung = future.result
                    individuals_by_rung[rung].append((loss, individual))
                    # Due to `evaluate` returning additional information (like the rung),
                    # evaluations are not automatically logged, so we do it here.
                    log_event(log, ASHA_LOG_TOKEN, rung, individual.fitness.wallclock_time,
                              individual.fitness.values, individual._id, individual.pipeline_str())
                    if rung == max(rungs):
                        log_event(log, TOKENS.EVALUATION_RESULT, individual.fitness.start_time,
                                  individual.fitness.wallclock_time, individual.fitness.process_time,
                                  individual.fitness.values, individual._id, individual.pipeline_str())

                start_new_job()

            highest_rung_reached = max(rungs)
    except stopit.TimeoutException:
        log.info('ASHA ended due to timeout.')
        highest_rung_reached = max(rung for rung, individuals in individuals_by_rung.items() if individuals != [])
        if highest_rung_reached != max(rungs):
            raise RuntimeWarning("Highest rung not reached.")
    finally:
        for rung, individuals in individuals_by_rung.items():
            log.info('[{}] {}'.format(rung, len(individuals)))

        return list(map(lambda p: p[1], individuals_by_rung[highest_rung_reached]))


def evaluate_on_rung(individual, rung, evaluate_individual, *args, **kwargs):
    individual = evaluate_individual(individual, *args, **kwargs)
    return individual, individual.fitness.values[0], rung



import logging
from gama.logging import MACHINE_LOG_LEVEL


class MachineLogFileHandler(logging.FileHandler):
    def __init__(self, filename, log_levels=None, mode='a', encoding=None, delay=False):
        """ As logging.FileHandler except only logs messages from `log_levels` to file.
        Used to log debug and error but not info.
        """
        super().__init__(filename, mode, encoding, delay)
        if log_levels is None:
            self.allowed_levels = [MACHINE_LOG_LEVEL, logging.WARNING, logging.ERROR, logging.CRITICAL]
        else:
            self.allowed_levels = log_levels
        super().setLevel(MACHINE_LOG_LEVEL)

    def emit(self, record):
        if record.levelno not in self.allowed_levels:
            return
        super().emit(record)



from contextlib import contextmanager
from typing import Iterator, Optional, NamedTuple, List, Any
import logging

from gama.logging.machine_logging import log_event, TOKENS
from .stopwatch import Stopwatch

log = logging.getLogger(__name__)


class Activity(NamedTuple):
    name: str
    stopwatch: Stopwatch
    time_limit: Optional[int] = None
    
    @property
    def time_left(self) -> float:
        """ Time left in seconds. Raises a TypeError if `time_limit` was not specified. """
        return self.time_limit - self.stopwatch.elapsed_time

    @property
    def exceeded_limit(self) -> float:
        """ True iff a limit was specified and its exceeded. False iff there is time left or no limit was specified. """
        return (self.time_limit is not None) and (self.time_limit - self.stopwatch.elapsed_time < 0)


class TimeKeeper:
    """ Simple object that helps keep track of time over multiple activities. """

    def __init__(self, total_time: Optional[int] = None):
        """
        Parameters
        ----------
        total_time: int, optional (default=None)
            The total time available across activities.
            If set to None, the `total_time_remaining` property will be unavailable.

        """
        self.total_time = total_time
        self.current_activity = None
        self.activities = []

    @property
    def total_time_remaining(self) -> float:
        """ Return time remaining in seconds. """
        if self.total_time is not None:
            return self.total_time - sum(map(lambda a: a.stopwatch.elapsed_time, self.activities))
        raise RuntimeError("Time Remaining only available if `total_time` was set on init.")

    @property
    def current_activity_time_elapsed(self) -> float:
        """ Return elapsed time in seconds of current activity. Raise RuntimeError if no current activity. """
        if self.current_activity is not None:
            return self.current_activity.stopwatch.elapsed_time
        else:
            raise RuntimeError("No activity in progress.")

    @property
    def current_activity_time_left(self) -> float:
        """ Return time left in seconds of current activity. Raise RuntimeError if no current activity. """
        if self.current_activity is not None and self.current_activity.time_limit is not None:
            return self.current_activity.time_limit - self.current_activity.stopwatch.elapsed_time
        elif self.current_activity is None:
            raise RuntimeError("No activity in progress.")
        else:
            raise RuntimeError("No time limit set for current activity.")

    @contextmanager
    def start_activity(self,
                       activity: str,
                       time_limit: Optional[int] = None,
                       activity_meta: Optional[List[Any]] = None) -> Iterator[Stopwatch]:
        """ Mark the start of a new activity and automatically time its duration.
            TimeManager does not currently support nested activities.

        Parameters
        ----------
        activity: str
            Name of the activity for reference in current activity or later look-ups.
        time_limit: int, optional (default=None)
            Intended time limit of the activity in seconds. Used to calculate time remaining.
        activity_meta: List[Any], optional (default=None)
            Any additional information about the activity to be logged.

        Returns
        -------
        ContextManager
            A context manager which when exited notes the end of the started activity.
        """
        if activity_meta is None:
            activity_meta = []
        log_event(log, TOKENS.PHASE_START, activity, *activity_meta)

        with Stopwatch() as sw:
            self.current_activity = Activity(activity, sw, time_limit)
            self.activities.append(self.current_activity)
            yield sw
        self.current_activity = None

        log_event(log, TOKENS.PHASE_END, activity, *activity_meta)
        log.info("{} took {:.4f}s.".format(activity, sw.elapsed_time))



""" Module to facility logging in machine parsable format. """
from datetime import datetime
from typing import List

from gama.logging import TIME_FORMAT, MACHINE_LOG_LEVEL

PLE_START = 'PLE'
PLE_END = 'END!'
PLE_DELIM = ';'


class TOKENS:
    EVALUATION_RESULT = 'EVAL'
    EVALUATION_ERROR = 'EVAL_ERR'
    PHASE_START = 'PS'
    PHASE_END = 'PE'
    EA_RESTART = 'EA_RST'
    EA_REMOVE_IND = 'RMV_IND'
    EVALUATION_TIMEOUT = 'EVAL_TO'
    MUTATION = 'IND_MUT'
    CROSSOVER = "IND_CX"
    INIT = 'INIT'

    @classmethod
    def values(cls) -> List[str]:
        """ Return a list of each possible TOKENS value. """
        # skips default dunder attributes like __module__
        return [getattr(cls, attribute) for attribute in vars(cls)
                if not (attribute.startswith('__') and attribute.endswith('__'))]


METHOD_TOKENS = dict(
    AsynchronousSuccessiveHalving='ASHA'
)


def default_time_format(datetime_: datetime):
    return datetime_.strftime(TIME_FORMAT)#[:-3]


def log_event(log_: object, token: object, *args: object) -> object:
    """ Writes the described event to the machine log level formatted for later parsing. """
    args = [default_time_format(arg) if isinstance(arg, datetime) else arg for arg in args]
    attrs = f'{PLE_DELIM}'.join([str(arg) for arg in args]) if not isinstance(args, str) else args
    message = f'{PLE_DELIM}'.join([PLE_START, token, attrs, default_time_format(datetime.now()), PLE_END])
    log_.log(level=MACHINE_LOG_LEVEL, msg=message)



import inspect
from typing import Union

import numpy as np
import pandas as pd
from sklearn.base import ClassifierMixin
from sklearn.preprocessing import LabelEncoder

from .gama import Gama
from gama.data import X_y_from_arff
from gama.configuration.classification import clf_config
from gama.utilities.metrics import scoring_to_metric


class GamaClassifier(Gama):
    """ Wrapper for the toolbox logic executing the AutoML pipeline for (multi-class) classification. """
    def __init__(self, config=None, scoring='neg_log_loss', *args, **kwargs):
        if not config:
            # Do this to avoid the whole dictionary being included in the documentation.
            config = clf_config

        self._metrics = scoring_to_metric(scoring)
        if any(metric.requires_probabilities for metric in self._metrics):
            # we don't want classifiers that do not have `predict_proba`, because then we have to
            # start doing one hot encodings of predictions etc.
            config = {alg: hp for (alg, hp) in config.items()
                      if not (inspect.isclass(alg) and issubclass(alg, ClassifierMixin)and not hasattr(alg(), 'predict_proba'))}

        self._label_encoder = None
        super().__init__(*args, **kwargs, config=config, scoring=scoring)

    def _predict(self, x: pd.DataFrame):
        """ Predict the target for input X.

        Parameters
        ----------
        x: pandas.DataFrame
            A dataframe with the same number of columns as the input to `fit`.

        Returns
        -------
        numpy.ndarray
            array with predictions of shape (N,) where N is the length of the first dimension of X.
        """
        y = self.model.predict(x)
        # Decode the predicted labels - necessary only if ensemble is not used.
        if y[0] not in self._label_encoder.classes_:
            y = self._label_encoder.inverse_transform(y)
        return y

    def _predict_proba(self, x: pd.DataFrame):
        """ Predict the class probabilities for input x.

        Predict target for x, using the best found pipeline(s) during the `fit` call.

        Parameters
        ----------
        x: pandas.DataFrame
            A dataframe with the same number of columns as the input to `fit`.

        Returns
        -------
        numpy.ndarray
            Array of shape (N, K) with class probabilities where N is the length of the
            first dimension of x, and K is the number of class labels found in `y` of `fit`.
        """
        return self.model.predict_proba(x)

    def predict_proba(self, x: Union[pd.DataFrame, np.ndarray]):
        """ Predict the class probabilities for input x.

        Predict target for x, using the best found pipeline(s) during the `fit` call.

        Parameters
        ----------
        x: pandas.DataFrame or numpy.ndarray
            A dataframe or numpy array with the same number of columns as the input to `fit`.

        Returns
        -------
        numpy.ndarray
            Array of shape (N, K) with class probabilities where N is the length of the
            first dimension of x, and K is the number of class labels found in `y` of `fit`.
        """
        if isinstance(x, np.ndarray):
            x = pd.DataFrame(x)
            for col in self._X.columns:
                x[col] = x[col].astype(self._X[col].dtype)
        return self._predict_proba(x)

    def predict_proba_arff(self, arff_file_path: str):
        """ Predict the class probabilities for input in the arff_file, must have empty target column.

        Predict target for X, using the best found pipeline(s) during the `fit` call.

        :param arff_file_path: str

        :return: a numpy array with class probabilities. The array is of shape (N, K) where N is the length of the
            first dimension of X, and K is the number of class labels found in `y` of `fit`.
        """
        X, _ = X_y_from_arff(arff_file_path)
        return self._predict_proba(X)

    def fit(self, x, y, *args, **kwargs):
        """ Should use base class documentation. """
        y_ = y.squeeze() if isinstance(y, pd.DataFrame) else y
        self._label_encoder = LabelEncoder().fit(y_)
        if any([isinstance(yi, str) for yi in y_]):
            # If target values are `str` we encode them or scikit-learn will complain.
            y = self._label_encoder.transform(y_)
        super().fit(x, y, *args, **kwargs)

    def _encode_labels(self, y):
        self._label_encoder = LabelEncoder().fit(y)
        return self._label_encoder.transform(y)



import itertools
import pandas as pd
from gama.utilities.preprocessing import format_x_y


def test_format_x_y():
    """ Tests that X and y data correctly get converted to (pd.DataFrame, pd.DataFrame). """
    def well_formatted_x_y(x, y, y_type):
        assert isinstance(x, pd.DataFrame)
        assert isinstance(y, y_type)
        assert len(x) == len(y)

    from sklearn.datasets import load_digits
    X_np, y_np = load_digits(return_X_y=True)
    X_df, y_df = pd.DataFrame(X_np), pd.DataFrame(y_np)
    y_series = pd.Series(y_np)

    for X, y in itertools.product([X_np, X_df], [y_np, y_series, y_df]):
        well_formatted_x_y(*format_x_y(X, y), y_type=pd.Series)
        well_formatted_x_y(*format_x_y(X, y, y_type=pd.DataFrame), y_type=pd.DataFrame)



from collections import namedtuple
import os
import pickle
import logging
import time
from typing import Optional, List

import pandas as pd
import numpy as np
from sklearn.preprocessing import OneHotEncoder
import stopit

from gama.genetic_programming.components import Individual
from gama.postprocessing.base_post_processing import BasePostProcessing
from gama.utilities.metrics import Metric, MetricType
from gama.utilities.generic.async_evaluator import AsyncEvaluator

log = logging.getLogger(__name__)
Model = namedtuple("Model", ['name', 'pipeline', 'predictions', 'validation_score'])


class EnsemblePostProcessing(BasePostProcessing):

    def __init__(self, time_fraction: float = 0.3, ensemble_size: Optional[int] = 25):
        """ Ensemble construction per Caruana et al.

        Parameters
        ----------
        time_fraction: float (default=0.3)
            Fraction of total time reserved for Ensemble building.
        ensemble_size: int, optional (default=25)
            Total number of models in the ensemble.
            When a single model is chosen more than once, it will increase its weight in the ensemble and
            *does* count towards this maximum.
        """
        super().__init__(time_fraction)
        self._hyperparameters = dict(
            ensemble_size=(ensemble_size, 25),
            metric=(None, None),
            cache=(None, None)
        )

    def dynamic_defaults(self, gama: 'Gama'):
        self._overwrite_hyperparameter_default('metric', gama._metrics[0])
        self._overwrite_hyperparameter_default('cache', gama._cache_dir)

    def post_process(self, x: pd.DataFrame, y: pd.Series, timeout: float, selection: List[Individual]) -> 'model':
        return build_fit_ensemble(x, y, timeout,
                                  self.hyperparameters['ensemble_size'],
                                  self.hyperparameters['metric'],
                                  self.hyperparameters['cache'])


class Ensemble(object):

    def __init__(self, metric, y: pd.DataFrame,
                 model_library=None, model_library_directory=None,
                 shrink_on_pickle=True):
        """
        Either model_library or model_library_directory must be specified.
        If model_library is specified, model_library_directory is ignored.

        Parameters
        ----------
        metric: string or Metric
            Metric to optimize the ensemble towards.
        y: pandas.DataFrame
            True labels for the predictions made by the models in the library.
        model_library:
            A list of models from which an ensemble can be built.
        model_library_directory: str
            a directory containing results of model evaluations.
        shrink_on_pickle: bool (default=True)
            If True, remove memory-intensive attributes that are required before pickling.
            When unpickled, the model can be used to create predictions, but the ensemble can't be changed.
        """
        if isinstance(metric, str):
            metric = Metric(metric)
        elif not isinstance(metric, Metric):
            raise ValueError("metric must be specified as string or `gama.ea.metrics.Metric`.")

        if model_library is None and model_library_directory is None:
            raise ValueError("At least one of model_library or model_library_directory must be specified.")

        if model_library is not None and model_library_directory is not None:
            log.warning("model_library_directory will be ignored because model_library is also specified.")

        if not isinstance(y, (pd.Series, pd.DataFrame)):
            raise TypeError(f"`y_true` must be of type pandas.DataFrame or pandas.Series but is {type(y)}.")

        self._metric = metric
        self._model_library_directory = model_library_directory
        self._model_library = model_library if model_library is not None else []
        self._shrink_on_pickle = shrink_on_pickle
        self._y = y
        self._prediction_transformation = None

        self._fit_models = None
        self._maximize = True
        self._child_ensembles = []
        self._models = {}

    @property
    def model_library(self):
        if not self._model_library:
            log.debug("Loading model library from disk.")
            self._model_library = load_predictions(self._model_library_directory, self._prediction_transformation)
            log.info("Loaded model library of size {} from disk.".format(len(self._model_library)))

        return self._model_library

    def _ensemble_validation_score(self, prediction_to_validate=None):
        raise NotImplementedError("Must be implemented by child class.")

    def _total_fit_weights(self):
        return sum([weight for (model, weight) in self._fit_models])

    def _total_model_weights(self):
        return sum([weight for (model, weight) in self._models.values()])

    def _averaged_validation_predictions(self):
        """ Get weighted average of predictions from the self._models on the hillclimb/validation set. """
        weighted_sum_predictions = sum([model.predictions * weight for (model, weight) in self._models.values()])
        return weighted_sum_predictions / self._total_model_weights()

    def build_initial_ensemble(self, n: int):
        """ Builds an ensemble of n models, based solely on the performance of individual models, not their combined performance.

        Parameters
        ----------
        n: int
            Number of models to include.
        """
        if not n > 0:
            raise ValueError("Ensemble must include at least one model.")
        if self._models:
            log.warning("The ensemble already contained models. Overwriting the ensemble.")
            self._models = {}

        sorted_ensembles = reversed(sorted(self.model_library, key=lambda m: m.validation_score))

        # Since the model library only features unique models, we do not need to check for duplicates here.
        selected_models = list(sorted_ensembles)[:n]
        for model in selected_models:
            self._add_model(model)

        log.debug("Initial ensemble created with score {}".format(self._ensemble_validation_score()))

    def _add_model(self, model, add_weight=1):
        """ Adds a specific model to the ensemble or increases its weight if it already is contained. """
        model, weight = self._models.pop(model.pipeline, (model, 0))
        self._models[model.pipeline] = (model, weight + add_weight)
        log.debug("Assigned a weight of {} to model {}".format(weight + add_weight, model.name))

    def expand_ensemble(self, n: int):
        """ Adds new models to the ensemble based on earlier given data.

        Parameters
        ----------
        n: int
            Number of models to add to current ensemble.
        """
        if not n > 0:
            raise ValueError("n must be greater than 0.")

        for _ in range(n):
            best_addition_score = -float('inf')
            current_weighted_average = self._averaged_validation_predictions()
            current_total_weight = self._total_model_weights()
            for model in self.model_library:
                if model.validation_score == 0:
                    continue
                candidate_pred = current_weighted_average + \
                                 (model.predictions - current_weighted_average) / (current_total_weight + 1)
                candidate_ensemble_score = self._ensemble_validation_score(candidate_pred)
                if best_addition_score < candidate_ensemble_score:
                    best_addition, best_addition_score = model, candidate_ensemble_score

            self._add_model(best_addition)
            log.debug('Ensemble size {} , best score: {}'.format(self._total_model_weights(), best_addition_score))

    def fit(self, X, y, timeout=1e6):
        """ Constructs an Ensemble out of the library of models.

        Parameters
        ----------
        X:
            Data to fit the final selection of models on.
        y:
            Targets corresponding to features X.
        timeout: int (default=1e6)
            Maximum amount of time in seconds that is allowed in total for fitting pipelines.
            If this time is exceeded, only pipelines fit until that point are taken into account when making
            predictions. Starting the parallelization takes roughly 4 seconds by itself.
        """
        if not self._models:
            raise RuntimeError("You need to call `build` to select models for the ensemble, before fitting them.")
        if timeout <= 0:
            raise ValueError("timeout must be greater than 0.")

        self._fit_models = []
        futures = set()
        with stopit.ThreadingTimeout(timeout) as c_mgr, AsyncEvaluator() as async_:
            for (model, weight) in self._models.values():
                futures.add(async_.submit(fit_and_weight, (model.pipeline, X, y, weight)))

            for _ in self._models.values():
                future = async_.wait_next()
                pipeline, weight = future.result
                if weight > 0:
                    self._fit_models.append((pipeline, weight))

        if not c_mgr:
            log.info("Fitting of ensemble stopped early.")

    def _get_weighted_mean_predictions(self, X, predict_method='predict'):
        weighted_predictions = []
        for (model, weight) in self._fit_models:
            target_prediction = getattr(model, predict_method)(X)
            if self._prediction_transformation:
                target_prediction = self._prediction_transformation(target_prediction)
            weighted_predictions.append(target_prediction * weight)

        return sum(weighted_predictions) / self._total_fit_weights()

    def __str__(self):
        if not self._models:
            return "Ensemble with no models."
        ensemble_str = "Ensemble of {} unique pipelines.\nR\tW\tScore\tPipeline\n".format(len(self._models))
        for i, (model, weight) in enumerate(sorted(self._models.values(), key=lambda x: x[0].validation_score)):
            ensemble_str += "{}\t{}\t{:.4f}\t{}\n".format(i, weight, model.validation_score[0], model.name)
        return ensemble_str

    def __getstate__(self):
        if self._shrink_on_pickle:
            log.info('Shrinking before pickle because shrink_on_pickle is True.'
                     'Removing anything that is not needed for predict-functionality.'
                     'Functionality to expand ensemble after unpickle is not available.')
            self._models = None
            self._model_library = None
            self._child_ensembles = None
            # self._y_true can not be removed as it is needed to ensure proper dimensionality of predictions
            # alternatively, one could just save the number of classes instead.

        return self.__dict__.copy()


def load_predictions(cache_dir, prediction_transformation=None):
    models = []
    for file in os.listdir(cache_dir):
        if file.endswith('.pkl'):
            file_name = os.path.join(cache_dir, file)
            if os.stat(file_name).st_size > 0:
                # We check file size, because writing to disk may be interrupted if the process was terminated due
                # to a restart/timeout. I can not find specifications saying that any interrupt of pickle.dump leads
                # to 0-sized files, but in practice this seems to case so far.
                with open(os.path.join(cache_dir, file), 'rb') as fh:
                    pl, predictions, scores = pickle.load(fh)
                if prediction_transformation:
                    predictions = prediction_transformation(predictions)
                models.append(Model(str(pl), pl, predictions, scores))
    return models


def fit_and_weight(args):
    """ Fit the pipeline given the data. Update weight to 0 if fitting fails.

    Parameters
    ----------
    args: Tuple
        Expected Tuple of [Pipeline, X, y, weight].

    Returns
    -------
    pipeline
        The same pipeline that was provided as input.
    weight
        If fitting succeeded, return the input weight.
        If *any* exception occurred during fitting, weight is 0.
    """
    pipeline, X, y, weight = args
    try:
        pipeline.fit(X, y)
    except Exception:
        log.warning("Exception when fitting pipeline {} of the ensemble. Assigning weight of 0."
                    .format(pipeline), exc_info=True)
        weight = 0

    return pipeline, weight


class EnsembleClassifier(Ensemble):
    def __init__(self, metric, y_true, label_encoder=None, *args, **kwargs):
        super().__init__(metric, y_true, *args, **kwargs)
        self._label_encoder = label_encoder

        # For metrics that only require class labels, we still want to apply one-hot-encoding to average predictions.
        y_as_squeezed_array = y_true.values.reshape(-1, 1)
        self._one_hot_encoder = OneHotEncoder(categories='auto').fit(y_as_squeezed_array)

        if self._metric.requires_probabilities:
            self._y = self._one_hot_encoder.transform(y_as_squeezed_array).toarray()
        else:
            def one_hot_encode_predictions(predictions):
                return self._one_hot_encoder.transform(predictions.reshape(-1, 1))

            self._prediction_transformation = one_hot_encode_predictions

    def _ensemble_validation_score(self, prediction_to_validate=None):
        if prediction_to_validate is None:
            prediction_to_validate = self._averaged_validation_predictions()

        if self._metric.requires_probabilities:
            return self._metric.maximizable_score(self._y, prediction_to_validate)
        else:
            # argmax returns (N, 1) matrix, need to squeeze it to (N,) for scoring.
            class_predictions = np.argmax(prediction_to_validate.toarray(), axis=1)
            return self._metric.maximizable_score(self._y, class_predictions)

    def predict(self, X):
        if self._metric.requires_probabilities:
            log.warning('Ensemble was tuned with a class-probabilities metric. '
                        'Using argmax of probabilities, which may not give optimal predictions.')
            class_probabilities = self._get_weighted_mean_predictions(X, 'predict_proba')
        else:
            class_probabilities = self._get_weighted_mean_predictions(X, 'predict').toarray()

        class_predictions = np.argmax(class_probabilities, axis=1)
        if self._label_encoder:
            class_predictions = self._label_encoder.inverse_transform(class_predictions)
        return class_predictions

    def predict_proba(self, X):
        if self._metric.requires_probabilities:
            return self._get_weighted_mean_predictions(X, 'predict_proba')
        else:
            log.warning('Ensemble was tuned with a class label predictions metric, not probabilities. '
                        'Using weighted mean of class predictions.')
            return self._get_weighted_mean_predictions(X, 'predict').toarray()


class EnsembleRegressor(Ensemble):
    def _ensemble_validation_score(self, prediction_to_validate=None):
        if prediction_to_validate is None:
            prediction_to_validate = self._averaged_validation_predictions()
        return self._metric.maximizable_score(self._y, prediction_to_validate)

    def predict(self, X):
        return self._get_weighted_mean_predictions(X)


def build_fit_ensemble(x, y, timeout: float, ensemble_size: int,
                       metric: Metric, cache: str, encoder: Optional[object]=None) -> Ensemble:
    """ Construct an Ensemble of models from cache, optimizing for metric and fit to (x, y). """
    start_build = time.time()

    log.debug('Building ensemble.')
    if metric.task_type == MetricType.CLASSIFICATION:
        ensemble = EnsembleClassifier(metric, y, model_library_directory=cache)
        ensemble._label_encoder = encoder
    elif metric.task_type == MetricType.REGRESSION:
        ensemble = EnsembleRegressor(metric, y, model_library_directory=cache)
    else:
        raise ValueError(f"Unknown metric task type {metric.task_type}")

    try:
        # Starting with more models in the ensemble should help against overfitting, but depending on the total
        # ensemble size, it might leave too little room to calibrate the weights or add new models. So we have
        # some adaptive defaults (for now).
        if ensemble_size <= 10:
            ensemble.build_initial_ensemble(1)
        else:
            ensemble.build_initial_ensemble(10)

        remainder = ensemble_size - ensemble._total_model_weights()
        if remainder > 0:
            ensemble.expand_ensemble(remainder)

        build_time = time.time() - start_build
        timeout = timeout - build_time
        log.info(f'Building ensemble took {build_time}s. Fitting ensemble with timeout {timeout}s.')

        ensemble.fit(x, y, timeout)
    except Exception as e:
        log.warning(f"Error during auto ensemble: {e}", exc_info=True)

    return ensemble



import logging
from typing import Tuple, Union, Type, Optional
import category_encoders as ce
import numpy as np
import pandas as pd
from sklearn.impute import SimpleImputer

log = logging.getLogger(__name__)


def define_preprocessing_steps(X_df: pd.DataFrame,
                               max_extra_features_created: Optional[int] = None,
                               max_categories_for_one_hot: int = 10):
    """ Constructs imputation and categorical preprocessing steps for the dataframe, taking into account data types.

    Parameters
    ----------
    X_df: pandas.DataFrame
        Features of the data (i.e. without target label).
    max_extra_features_created: int, optional (default=None)
        WARNING! Currently not supported, the only valid value is None.

        If set, dynamically decrease max_categories_for_one_hot as needed so as not to
        create a greater amount of features than max_extra_features_created.
    max_categories_for_one_hot: int (default=10)
        The maximum amount of unique category levels to be considered for one hot encoding.

    Returns
    -------
    List
        List of preprocessing step objects.
    """
    if max_extra_features_created:
        # Will determine max_categories_for_one_hot based on how many total new features would be created.
        raise NotImplementedError()

    one_hot_columns = []
    target_encoding_columns = []
    for unique_values, dtype, column_index in zip(X_df.apply(pd.Series.nunique), X_df.dtypes, X_df.columns):
        if isinstance(dtype, pd.core.dtypes.dtypes.CategoricalDtype):
            if unique_values > max_categories_for_one_hot:
                target_encoding_columns.append(column_index)
            elif unique_values > 1:
                one_hot_columns.append(column_index)
            else:
                pass  # Binary category or constant feature.

    one_hot_encoder = ce.OneHotEncoder(cols=one_hot_columns, handle_unknown='ignore')
    target_encoder = ce.TargetEncoder(cols=target_encoding_columns, handle_unknown='ignore')
    imputer = SimpleImputer(strategy='median')

    return [one_hot_encoder, target_encoder, imputer]


def heuristic_numpy_to_dataframe(X: np.ndarray, max_unique_values_cat: int = 10) -> pd.DataFrame:
    """ Transform a numpy array to a typed pd.DataFrame. """
    X_df = pd.DataFrame(X)
    for column, n_unique in X_df.nunique(dropna=True).items():
        if n_unique <= max_unique_values_cat:
            X_df[column] = X_df[column].astype('category')
    return X_df


def format_x_y(x: Union[pd.DataFrame, np.ndarray], y: Union[pd.DataFrame, pd.Series, np.ndarray],
               y_type: Type=pd.Series, remove_unlabeled: bool = True
               ) -> Tuple[pd.DataFrame, Union[pd.DataFrame, pd.Series]]:
    """ Takes various types of (X,y) data and converts it into a (pd.DataFrame, pd.Series) tuple.

    Parameters
    ----------
    x: pandas.DataFrame or numpy.ndarray
    y: pandas.DataFrame, pandas.Series or numpy.ndarray
    y_type: Type (default=pandas.Series)
    remove_unlabeled: bool (default=True)
        If true, remove all rows associated with unlabeled data (NaN in y).

    Returns
    -------
    Tuple[pandas.DataFrame, pandas.DataFrame or pandas.Series]
        X and y, where X is formatted as pd.DataFrame and y is formatted as `y_type`.
    """
    if not isinstance(x, (np.ndarray, pd.DataFrame)):
        raise TypeError("X must be either np.ndarray or pd.DataFrame.")
    if not isinstance(y, (np.ndarray, pd.Series, pd.DataFrame)):
        raise TypeError("y must be np.ndarray, pd.Series or pd.DataFrame.")

    if isinstance(x, np.ndarray):
        x = heuristic_numpy_to_dataframe(x)
    if isinstance(y, np.ndarray) and y.ndim == 2 and y.shape[1] > 1:
        y = np.argmax(y, axis=1)

    if y_type == pd.Series:
        if isinstance(y, pd.DataFrame):
            y = y.squeeze()
        elif isinstance(y, np.ndarray):
            y = pd.Series(y)
    elif y_type == pd.DataFrame:
        if not isinstance(y, pd.DataFrame):
            y = pd.DataFrame(y)
    else:
        raise ValueError(f"`y_type` must be one of [pandas.Series, pandas.DataFrame] but is {y_type}.")

    if remove_unlabeled:
        unlabeled = y[y.columns[0]].isnull() if isinstance(y, pd.DataFrame) else y.isnull()
        if unlabeled.isnull().any():
            log.info("Target vector has been found to contain NaN-labels, these rows will be ignored.")
            x, y = x.loc[~y.isnull(), :], y[~y.isnull(), :]

    return x, y



"""
I don't want to be reinventing the wheel but I can't find a satisfying implementation.
I want to be able to execute arbitrary functions asynchronously on a different process.
Any ongoing subprocesses must immediately be able to be terminated without errors.
Results of cancelled subprocesses may be ignored.

`concurrent.futures.ProcessPoolExecutor` gets very close to the desired implementation,
but it has issues:
    - by default it waits for subprocesses to close on __exit__.
      Unfortunately it is possible the subprocesses can be running non-Python code,
      e.g. a C implementation of SVC, meaning the subprocess won't end until the SVC fit is complete.
    - even if that is overwritten and no wait is performed, the subprocess will throw an error when it is done.
      Though that does not hinder the execution of the program, I don't want errors for expected behavior.
"""
import itertools
import logging
import multiprocessing
import queue
import time
import uuid
from typing import Optional, Callable

log = logging.getLogger(__name__)


class AsyncFuture:
    """ Reference to a function call executed on a different process. """

    def __init__(self, fn, *args, **kwargs):
        self.id = uuid.uuid4()
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.result = None
        self.exception = None

    def execute(self):
        """ Execute the function call `fn(*args, **kwargs)` and record results. """
        try:
            self.result = self.fn(*self.args, **self.kwargs)
        except Exception as e:
            self.exception = e


class AsyncEvaluator:
    """ ContextManager which manages subprocesses on which arbitrary* functions can be evaluated.

    * The function and all its arguments must be picklable.

    You can not use the same AsyncEvaluator in two different contexts. Doing so raises a `RuntimeError`.
    """
    n_jobs: int = multiprocessing.cpu_count()

    def __init__(self, n_workers: Optional[int] = None):
        """
        Parameters
        ----------
        n_workers : int, optional (default=None)
            Maximum number of subprocesses to run for parallel evaluations.
            If None, use `AsyncEvaluator.n_jobs` which defaults to multiprocessing.cpu_count().
        """
        self._has_entered = False
        self.futures = {}
        self._processes = []
        self._n_jobs = n_workers if n_workers is not None else AsyncEvaluator.n_jobs

        self._queue_manager = multiprocessing.Manager()
        self._input_queue = self._queue_manager.Queue()
        self._output_queue = self._queue_manager.Queue()

    def __enter__(self):
        if self._has_entered:
            raise RuntimeError("You can not use the same AsyncEvaluator in two different contexts.")
        self._has_entered = True

        self._input_queue = self._queue_manager.Queue()
        self._output_queue = self._queue_manager.Queue()

        log.debug(f"Starting {self._n_jobs} subprocesses.")
        for _ in range(self._n_jobs):
            subprocess = multiprocessing.Process(
                target=evaluator_daemon,
                args=(self._input_queue, self._output_queue),
                daemon=True
            )
            self._processes.append(subprocess)
            subprocess.start()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        log.debug(f"Terminating {len(self._processes)} subprocesses.")
        # This is ugly as the subprocesses use shared queues and in direct conflict with guidelines:
        # https://docs.python.org/3/library/multiprocessing.html#all-start-methods
        for subprocess in self._processes:
            subprocess.terminate()
        return False

    def submit(self, fn: Callable, *args, **kwargs) -> AsyncFuture:
        """ Submit fn(*args, **kwargs) to be evaluated on a subprocess.

        Parameters
        ----------
        fn: Callable
            Function to call on a subprocess.
        args
            Positional arguments to call `fn` with.
        kwargs
            Keyword arguments to call `fn` with.

        Returns
        -------
        AsyncFuture
            A Future of which the `result` or `exception` field will be populated once evaluation is finished.
        """
        future = AsyncFuture(fn, *args, **kwargs)
        self.futures[future.id] = future
        self._input_queue.put(future)
        return future

    def wait_next(self, poll_time: float = 0.05) -> AsyncFuture:
        """ Wait until an AsyncFuture has been completed and return it.


        Parameters
        ----------
        poll_time: float (default=0.05)
            Time to sleep between checking if a future has been completed.

        Returns
        -------
        AsyncFuture
            The completed future that completed first.

        Raises
        ------
        RuntimeError
            If all futures have already been completed and returned.
        """
        if len(self.futures) == 0:
            raise RuntimeError("No Futures queued, must call `submit` first.")

        while True:
            try:
                completed_future = self._output_queue.get(block=False)
                matching_future = self.futures.pop(completed_future.id)
                matching_future.result, matching_future.exception = completed_future.result, completed_future.exception
                return matching_future
            except queue.Empty:
                time.sleep(poll_time)
                continue


def evaluator_daemon(
        input_queue: queue.Queue,
        output_queue: queue.Queue,
        print_exit_message: bool = True):
    """ Function for daemon subprocess that evaluates functions from AsyncFutures.

    Parameters
    ----------
    input_queue: queue.Queue[AsyncFuture]
        Queue to get AsyncFuture from. Queue should be managed by multiprocessing.manager.
    output_queue: queue.Queue[AsyncFuture]
        Queue to put AsyncFuture to. Queue should be managed by multiprocessing.manager.
    print_exit_message: bool (default=True)
        If True, print to console the reason for shutting down.
        If False, shut down silently.
    """
    try:
        while True:
            future = input_queue.get()
            future.execute()
            output_queue.put(future)
    except KeyboardInterrupt:
        shutdown_message = 'Helper process stopping due to keyboard interrupt.'
    except (BrokenPipeError, EOFError):
        shutdown_message = 'Helper process stopping due to a broken pipe or EOF.'
    if print_exit_message:
        print(shutdown_message)



""" Functions which take two Individuals, and produce at least one new Individual from them. """
import random
from typing import List, Callable, Iterable

from gama.genetic_programming.components import Individual


def random_crossover(individual1: Individual, individual2: Individual) -> None:
    """ Perform a random valid crossover between two individuals in-place, if there is one.

    Parameters
    ----------
    individual1: Individual
        The individual to crossover with individual2.
    individual2: Individual
        The individual to crossover with individual1.

    Raises
    ------
    ValueError
        If there is no valid crossover function for the two individuals.
    """
    crossover_choices = _valid_crossover_functions(individual1, individual2)
    if len(crossover_choices) == 0:
        raise ValueError(f"{individual1.pipeline_str()} and {individual2.pipeline_str()} can't mate.")
    random.choice(crossover_choices)(individual1, individual2)


def crossover_primitives(individual1: Individual, individual2: Individual) -> None:
    """ Crossover two individuals by splitting both at a random PrimitiveNode and switching one part out.

    Parameters
    ----------
    individual1: Individual
        The individual to crossover with individual2.
    individual2: Individual
        The individual to crossover with individual1.
    """
    parent_node_1 = random.choice(list(individual1.primitives)[:-1])
    parent_node_2 = random.choice(list(individual2.primitives)[:-1])
    parent_node_1._data_node, parent_node_2._data_node = parent_node_2._data_node, parent_node_1._data_node


def crossover_terminals(individual1: Individual, individual2: Individual) -> None:
    """ Crossover two individuals in-place by exchanging two Terminals with shared output type but different values.

    Parameters
    ----------
    individual1: Individual
        The individual to crossover with individual2.
    individual2: Individual
        The individual to crossover with individual1.
    """
    candidates = list(_shared_terminals(individual1, individual2, with_indices=True, value_match='different'))
    i, ind1_term, j, ind2_term = random.choice(candidates)
    individual1.replace_terminal(i, ind2_term)
    individual2.replace_terminal(j, ind1_term)


def _shared_terminals(individual1: Individual, individual2: Individual,
                      with_indices: bool = True, value_match: str = 'different') -> Iterable:
    """ Finds all shared Terminals between two Individuals.

    Parameters
    ----------
    individual1: Individual
    individual2: Individual
    with_indices: bool (default=True)
        If True, also return the indices of the Terminals w.r.t. the Individual.
    value_match: str (default='different')
        Indicates with matches to return, based on terminal values.
        Accepted values are:

         - 'different': only return shared terminals which have different values from each other
         - 'equal': only return shared terminals which have equal values from each other
         - 'all': return all shared terminals regardless of value

    Returns
    -------
    Sequence of Tuples with both Terminals, with the Terminal from Individual1 first.
        Tuple[Terminal, Terminal] if `with_indices` is False
        Tuple[int, Terminal, int, Terminal] if `with_indices` is True,
        each int specifies the index of the Terminal directly after.
    """
    if value_match not in ['different', 'equal', 'all']:
        raise ValueError(f"`value_match` must be one of 'all', 'equal' or 'different' but is '{value_match}'.")

    for i, ind1_term in enumerate(individual1.terminals):
        for j, ind2_term in enumerate(individual2.terminals):
            if ind1_term.identifier == ind2_term.identifier:
                if ((value_match == 'different' and ind1_term.value == ind2_term.value)
                        or (value_match == 'equal' and ind1_term.value != ind2_term.value)):
                    continue
                if with_indices:
                    yield (i, ind1_term, j, ind2_term)
                else:
                    yield (ind1_term, ind2_term)


def _valid_crossover_functions(individual1: Individual, individual2: Individual) -> List[Callable]:
    """ Return all crossover functions which can produce a new individual from the given two individuals.

    Parameters
    ----------
    individual1: Individual
        The individual to crossover with individual2.
    individual2: Individual
        The individual to crossover with individual1.

    Returns
    -------
    List[Callable]
        List of valid crossover functions given the content of individual1 and individual2.
    """
    crossover_choices = []
    if list(_shared_terminals(individual1, individual2)) != []:
        crossover_choices.append(crossover_terminals)
    if len(list(individual1.primitives)) >= 2 and len(list(individual2.primitives)) >= 2:
        crossover_choices.append(crossover_primitives)
    return crossover_choices



import logging
from typing import List, Optional, Callable

import pandas as pd

from gama.genetic_programming.components import Individual
from gama.genetic_programming.operator_set import OperatorSet
from gama.search_methods.base_search import BaseSearch, _check_base_search_hyperparameters
from gama.utilities.generic.async_evaluator import AsyncEvaluator

log = logging.getLogger(__name__)


class RandomSearch(BaseSearch):
    """ Perform random search over all possible pipelines. """
    def dynamic_defaults(self, x: pd.DataFrame, y: pd.DataFrame, time_limit: int):
        pass

    def search(self, operations: OperatorSet, start_candidates: List[Individual]):
        random_search(operations, self.output, start_candidates)


def random_search(
        operations: OperatorSet,
        output: List[Individual],
        start_candidates: List[Individual],
        max_evaluations: Optional[int] = None) -> List[Individual]:
    """ Perform random search over all possible pipelines.

    Parameters
    ----------
    operations: OperatorSet
        An operator set with `evaluate` and `individual` functions.
    output: List[Individual]
        A list which contains the found individuals during search.
    start_candidates: List[Individual]
        A list with candidate individuals to evaluate first.
    max_evaluations: int, optional (default=None)
        If specified, only a maximum of `max_evaluations` individuals are evaluated.
        If None, the algorithm will be run indefinitely.

    Returns
    -------
    List[Individual]
        All evaluated individuals.
    """
    _check_base_search_hyperparameters(operations, output, start_candidates)

    with AsyncEvaluator() as async_:
        for individual in start_candidates:
            async_.submit(operations.evaluate, individual)

        while (max_evaluations is None) or (len(output) < max_evaluations):
            future = operations.wait_next(async_)
            if future.result is not None:
                output.append(future.result)
            async_.submit(operations.evaluate, operations.individual())

    return output
